"""
TerminalCommand class is used to run commands from python as if we are in a shell/cmd
"""
import subprocess as p

__all__ = ['TerminalCommand']


class TerminalCommand:
    def __init__(self, cmd, cwd=None, env=None):
        """
        Creates a new command container.
        Note that this immediately executes the command synchronously and writes the
        return values to the corresponding members.

        Attributes
        ----------
        cmd : The command to execute.
        cwd : The working directory to run the command in.
        env : The environment the command is run in.
        """
        process = p.Popen(cmd, stdout=p.PIPE, stderr=p.PIPE, shell=True, cwd=cwd, env=env)
        stdout, stderr = process.communicate()

        self.stdout = stdout.decode('utf8')
        self.stderr = stderr.decode('utf8')

        self.return_code = process.returncode
        process.kill()

    def is_success(self):
        return self.return_code == 0

    def is_failure(self):
        return self.return_code != 0

    def get_split_output_strings(self):
        lines = self.stdout.splitlines()
        words = []
        [words.append(word) for line in lines for word in line.split()]
        return words


if __name__ == "__main__":
    pass



import logging
from compas.geometry import Vector, normalize_vector
from compas_slicer.geometry import PrintPoint
import compas_slicer.utilities as utils

logger = logging.getLogger('logger')

__all__ = ['BaseBoundary']


class BaseBoundary:
    """
    The BaseBoundary is like a fake initial layer that supports the first path of the segment.
    This is useful, because for our computations we need to have a support layer for evey path.
    The first path has as support the Base Boundary, and every other path has its previous path.

    Attributes
    ----------
    mesh :
    points :
    override_vector :
    """

    def __init__(self, mesh, points, override_vector=None):
        self.mesh = mesh
        self.points = points
        self.override_vector = override_vector
        if self.override_vector:
            self.up_vectors = [self.override_vector for p in self.points]
        else:
            self.up_vectors = self.get_up_vectors()

        self.printpoints = [PrintPoint(pt=pt,  # Create fake print points
                                       layer_height=1.0,
                                       mesh_normal=utils.get_closest_mesh_normal_to_pt(self.mesh, pt))
                            for pt in self.points]

        for i, pp in enumerate(self.printpoints):
            pp.up_vector = self.up_vectors[i]

    def __repr__(self):
        return "<BaseBoundary object with %i points>" % len(self.points)

    def get_up_vectors(self):
        """ Finds the up_vectors of each point of the boundary. A smoothing step is also included. """
        up_vectors = []
        for i, p in enumerate(self.points):
            v1 = Vector.from_start_end(p, self.points[(i + 1) % len(self.points)])
            cross = v1.cross(utils.get_closest_mesh_normal_to_pt(self.mesh, p))
            v = normalize_vector(cross)
            v = Vector(v[0], v[1], v[2])
            if v[2] < 0:
                v.scale(-1)
            up_vectors.append(v)
        up_vectors = utils.smooth_vectors(up_vectors, strength=0.4, iterations=3)
        return up_vectors

    def to_data(self):
        """ Returns a dictionary with the data of the class. """
        return {"points": utils.point_list_to_dict(self.points),
                "up_vectors": utils.point_list_to_dict(self.up_vectors)}


if __name__ == "__main__":
    pass



import logging
from compas_slicer.pre_processing.preprocessing_utils import blend_union_list

logger = logging.getLogger('logger')

__all__ = ['assign_interpolation_distance_to_mesh_vertices',
           'assign_interpolation_distance_to_mesh_vertex']


def assign_interpolation_distance_to_mesh_vertices(mesh, weight, target_LOW, target_HIGH):
    """
    Fills in the 'get_distance' attribute of every vertex of the mesh.

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    weight: float,
        The weighting of the distances from the lower and the upper target, from 0 to 1.
    target_LOW: :class: 'compas_slicer.pre_processing.CompoundTarget'
        The lower compound target.
    target_HIGH:  :class: 'compas_slicer.pre_processing.CompoundTarget'
        The upper compound target.
    """
    for i, vkey in enumerate(mesh.vertices()):
        d = assign_interpolation_distance_to_mesh_vertex(vkey, weight, target_LOW, target_HIGH)
        mesh.vertex[vkey]['scalar_field'] = d


def assign_interpolation_distance_to_mesh_vertex(vkey, weight, target_LOW, target_HIGH):
    """
    Fills in the 'get_distance' attribute for a single vertex with vkey.

    Parameters
    ----------
    vkey: int
        The vertex key.
    weight: float,
        The weighting of the distances from the lower and the upper target, from 0 to 1.
    target_LOW: :class: 'compas_slicer.pre_processing.CompoundTarget'
        The lower compound target.
    target_HIGH:  :class: 'compas_slicer.pre_processing.CompoundTarget'
        The upper compound target.
    """
    if target_LOW and target_HIGH:  # then interpolate targets
        d = get_weighted_distance(vkey, weight, target_LOW, target_HIGH)
    elif target_LOW:  # then offset target
        offset = weight * target_LOW.get_max_dist()
        d = target_LOW.get_distance(vkey) - offset
    else:
        raise ValueError('You need to provide at least one target')
    return d


def get_weighted_distance(vkey, weight, target_LOW, target_HIGH):
    """
    Computes the weighted get_distance for a single vertex with vkey.

    Parameters
    ----------
    vkey: int
        The vertex key.
    weight: float,
        The weighting of the distances from the lower and the upper target, from 0 to 1.
    target_LOW: :class: 'compas_slicer.pre_processing.CompoundTarget'
        The lower compound target.
    target_HIGH:  :class: 'compas_slicer.pre_processing.CompoundTarget'
        The upper compound target.
    """
    # --- calculation with uneven weights
    if target_HIGH.has_uneven_weights:
        d_low = target_LOW.get_distance(vkey)  # float
        ds_high = target_HIGH.get_all_distances_for_vkey(vkey)  # list of floats (# number_of_boundaries)

        if target_HIGH.number_of_boundaries > 1:
            weights_remapped = [remap_unbound(weight, 0, weight_max, 0, 1)
                                for weight_max in target_HIGH.weight_max_per_cluster]
            weights = weights_remapped
        else:
            weights = [weight]

        distances = [(weight - 1) * d_low + weight * d_high for d_high, weight in zip(ds_high, weights)]

        if target_HIGH.has_blend_union:
            return blend_union_list(distances, target_HIGH.blend_radius)
        else:
            return min(distances)

    # --- simple calculation (without uneven weights)
    else:
        d_low = target_LOW.get_distance(vkey)
        d_high = target_HIGH.get_distance(vkey)
        return (d_low * (1 - weight)) - (d_high * weight)


def remap_unbound(input_val, in_from, in_to, out_from, out_to):
    """
    Remaps input_val from source domain to target domain.
    No clamping is performed, the result can be outside of the target domain
    if the input is outside of the source domain.
    """
    out_range = out_to - out_from
    in_range = in_to - in_from
    in_val = input_val - in_from
    val = (float(in_val) / in_range) * out_range
    out_val = out_from + val
    return out_val


if __name__ == "__main__":
    pass



"""
********************************************************************************
parameters
********************************************************************************

.. autosummary::
    :toctree: generated/
    :nosignatures:

    get_param
    defaults_curved_slicing
    defaults_gcode
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

#  Polyline simplification
from .get_param import *  # noqa: F401 E402 F403
from .defaults_interpolation_slicing import *  # noqa: F401 E402 F403
from .defaults_gcode import *  # noqa: F401 E402 F403
from .defaults_layers import *  # noqa: F401 E402 F403
from .defaults_print_organization import *  # noqa: F401 E402 F403


__all__ = [name for name in dir() if not name.startswith('_')]



import numpy as np
import logging
import compas_slicer.utilities as utils
from compas.plugins import PluginNotInstalledError
from compas_slicer.pre_processing.preprocessing_utils.gradient import get_scalar_field_from_gradient, \
    get_face_gradient_from_scalar_field, normalize_gradient
import scipy
import math

packages = utils.TerminalCommand('conda list').get_split_output_strings()
if 'igl' in packages:
    import igl

logger = logging.getLogger('logger')

__all__ = ['get_igl_EXACT_geodesic_distances',
           'get_custom_HEAT_geodesic_distances']


def check_igl_is_installed():
    """ Throws an error if igl python bindings are not installed in the current environment. """
    if 'igl' not in packages:
        raise PluginNotInstalledError("--------ATTENTION! ----------- \
                        Libigl library is missing! \
                        Install it with 'conda install -c conda-forge igl'")


def get_igl_EXACT_geodesic_distances(mesh, vertices_start):
    """
    Calculate geodesic distances using libigl.

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    vertices_start: list, int
    """
    check_igl_is_installed()

    v, f = mesh.to_vertices_and_faces()
    v = np.array(v)
    f = np.array(f)
    vertices_target = np.arange(len(v))  # all vertices are targets
    vstart = np.array(vertices_start)
    distances = igl.exact_geodesic(v, f, vstart, vertices_target)
    return distances


def get_custom_HEAT_geodesic_distances(mesh, vi_sources, OUTPUT_PATH, v_equalize=None, anisotropic_scaling=False):
    """ Calculate geodesic distances using the heat method. """
    check_igl_is_installed()

    geodesics_solver = GeodesicsSolver(mesh, OUTPUT_PATH)
    u = geodesics_solver.diffuse_heat(vi_sources, v_equalize, method='simulation')
    geodesic_dist = geodesics_solver.get_geodesic_distances(u, vi_sources, v_equalize)
    return geodesic_dist


######################################
# --- GeodesicsSolver

USE_FORWARDS_EULER = False
HEAT_DIFFUSION_ITERATIONS = 250
DELTA = 0.1


class GeodesicsSolver:
    """
    Computes custom geodesic distances. Starts from implementation of the method presented in the paper
    'Geodesics in Heat' (Crane, 2013)

    Attributes
    ----------
    mesh: :class: compas.datastructures.Mesh
    OUTPUT_PATH: str
    """

    def __init__(self, mesh, OUTPUT_PATH):
        logger.info('GeodesicsSolver')
        self.mesh = mesh
        self.OUTPUT_PATH = OUTPUT_PATH

        self.use_forwards_euler = True

        v, f = mesh.to_vertices_and_faces()
        v = np.array(v)
        f = np.array(f)

        # compute necessary data
        self.cotans = igl.cotmatrix_entries(v, f)  # compute_cotan_field(self.mesh)
        self.L = igl.cotmatrix(v, f)  # assemble_laplacian_matrix(self.mesh, self.cotans)
        self.M = igl.massmatrix(v, f)  # create_mass_matrix(mesh)

    def diffuse_heat(self, vi_sources, v_equalize=None, method='simulation'):
        """
        Heat diffusion.

        Attributes
        ----------
        vi_sources: list, int, the vertex indices of the sources
        v_equalize: list, int, the vertex indices whose value should be equalized
        method: str (Currently only 'simulation' works.)
        """
        if not v_equalize:
            v_equalize = []

        # First assign starting values (0 everywhere, 1 on the sources)
        u0 = np.zeros(len(list(self.mesh.vertices())))
        u0[vi_sources] = 1.0
        u = u0

        if method == 'default':  # This is buggy, does not keep boundary exactly on 0. TODO: INVESTIGATE
            t_mult = 1
            t = t_mult * np.mean(np.array([self.mesh.face_area(fkey) for fkey in self.mesh.faces()]))  # avg face area
            solver = scipy.sparse.linalg.factorized(self.M - t * self.L)  # pre-factor solver
            u = solver(u0)  # solve the heat equation: u = (VA - t * Lc) * u0

        elif method == 'simulation':
            u = u0

            for i in range(HEAT_DIFFUSION_ITERATIONS):
                if USE_FORWARDS_EULER:  # Forwards Euler (doesn't work so well)
                    u_prime = u + DELTA * self.L * u
                else:  # Backwards Euler
                    #  (M-delta*L) * u_prime = M*U
                    S = (self.M - DELTA * self.L)
                    b = self.M * u
                    u_prime = scipy.sparse.linalg.spsolve(S, b)

                if len(v_equalize) > 0:
                    u_prime[v_equalize] = np.min(u_prime[v_equalize])

                u = u_prime
                u[vi_sources] = 1.0  # make sure sources remain fixed to 1

        # reverse values (to make vstarts on 0)
        u = ([np.max(u)] * len(u)) - u

        utils.save_to_json([float(value) for value in u], self.OUTPUT_PATH, 'diffused_heat.json')
        return u

    def get_geodesic_distances(self, u, vi_sources, v_equalize=None):
        """
        Finds geodesic distances from heat distribution u. I

        Parameters
        ----------
        u: np.array, dimensions: V x 1 (one scalar value per vertex)
        vi_sources: list, int, the vertex indices of the sources
        v_equalize: list, int, the vertex indices whose value should be equalized
        """
        X = get_face_gradient_from_scalar_field(self.mesh, u)
        X = normalize_gradient(X)
        geodesic_dist = get_scalar_field_from_gradient(self.mesh, X, self.L, self.cotans)
        assert not math.isnan(geodesic_dist[0]), \
            "Attention, the 'get_scalar_field_from_gradient' function returned Nan. "
        geodesic_dist[vi_sources] = 0  # coerce boundary vertices to be on 0 (fixes small boundary imprecision)
        return geodesic_dist


if __name__ == "__main__":
    pass



import os
import json
import logging
from compas.geometry import Point, distance_point_point_sqrd, normalize_vector
from compas.geometry import Vector, length_vector, closest_point_in_cloud
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import scipy

logger = logging.getLogger('logger')

__all__ = ['get_output_directory',
           'save_to_json',
           'load_from_json',
           'is_jsonable',
           'save_to_text_file',
           'flattened_list_of_dictionary',
           'interrupt',
           'point_list_to_dict',
           'point_list_from_dict',
           'get_closest_mesh_vkey_to_pt',
           'get_closest_mesh_normal_to_pt',
           'get_mesh_cotmatrix_igl',
           'get_mesh_cotans_igl',
           'get_closest_pt_index',
           'get_closest_pt',
           'plot_networkx_graph',
           'get_mesh_vertex_coords_with_attribute',
           'get_dict_key_from_value',
           'find_next_printpoint',
           'find_previous_printpoint',
           'smooth_vectors',
           'get_normal_of_path_on_xy_plane',
           'get_all_files_with_name']


def get_output_directory(path):
    """
    Checks if a directory with the name 'output' exists in the path. If not it creates it.

    Parameters
    ----------
    path: str
        The path where the 'output' directory will be created

    Returns
    ----------
    str
        The path to the new (or already existing) 'output' directory
    """
    output_dir = os.path.join(path, 'output')
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)
    return output_dir


def get_closest_pt_index(pt, pts):
    """
    Finds the index of the closest point of 'pt' in the point cloud 'pts'.

    Parameters
    ----------
    pt: compas.geometry.Point3d
    pts: list, compas.geometry.Point3d

    Returns
    ----------
    int
        The index of the closest point
    """
    ci = closest_point_in_cloud(point=pt, cloud=pts)[2]
    # distances = [distance_point_point_sqrd(p, pt) for p in pts]
    # ci = distances.index(min(distances))
    return ci


def get_closest_pt(pt, pts):
    """
     Finds the closest point of 'pt' in the point cloud 'pts'.

    Parameters
    ----------
    pt: :class: 'compas.geometry.Point'
    pts: list, :class: 'compas.geometry.Point3d'

    Returns
    ----------
    compas.geometry.Point3d
        The closest point
    """
    ci = closest_point_in_cloud(point=pt, cloud=pts)[2]
    return pts[ci]


def smooth_vectors(vectors, strength, iterations):
    """
    Smooths the vector iteratively, with the given number of iterations and strength per iteration

    Parameters
    ----------
    vectors: list, :class: 'compas.geometry.Vector'
    strength: float
    iterations: int

    Returns
    ----------
    list, :class: 'compas.geometry.Vector3d'
        The smoothened vectors
    """

    for _ in range(iterations):
        for i, n in enumerate(vectors):
            if 0 < i < len(vectors) - 1:
                neighbors_average = (vectors[i - 1] + vectors[i + 1]) * 0.5
            else:
                neighbors_average = n
            vectors[i] = n * (1 - strength) + neighbors_average * strength
    return vectors


#######################################
#  json

def save_to_json(data, filepath, name):
    """
    Save the provided data to json on the filepath, with the given name

    Parameters
    ----------
    data: dict_or_list
    filepath: str
    name: str
    """

    filename = os.path.join(filepath, name)
    logger.info("Saving to json: " + filename)
    with open(filename, 'w') as f:
        f.write(json.dumps(data, indent=3, sort_keys=True))


def load_from_json(filepath, name):
    """
    Loads json from the filepath

    Parameters
    ----------
    filepath: str
    name: str
    """

    filename = os.path.join(filepath, name)
    with open(filename, 'r') as f:
        data = json.load(f)
    logger.info("Loaded json: " + filename)
    return data


def is_jsonable(x):
    """ Returns True if x can be json-serialized, False otherwise. """
    try:
        json.dumps(x)
        return True
    except TypeError:
        return False


#######################################
#  text file

def save_to_text_file(data, filepath, name):
    """
    Save the provided text on the filepath, with the given name

    Parameters
    ----------
    data: str
    filepath: str
    name: str
    """

    filename = os.path.join(filepath, name)
    logger.info("Saving to text file: " + filename)
    with open(filename, 'w') as f:
        f.write(data)


#######################################
#  mesh utils

def check_triangular_mesh(mesh):
    """
    Checks if the mesh is triangular. If not, then it raises an error

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    """

    for f_key in mesh.faces():
        vs = mesh.face_vertices(f_key)
        if len(vs) != 3:
            raise TypeError("Found a quad at face key: " + str(f_key) + " ,number of face vertices:" + str(
                len(vs)) + ". \nOnly triangular meshes supported.")


def get_closest_mesh_vkey_to_pt(mesh, pt):
    """
    Finds the vertex key that is the closest to the point.

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    pt: :class: 'compas.geometry.Point'

    Returns
    ----------
    int
        the closest vertex key
    """
    # cloud = [Point(data['x'], data['y'], data['z']) for v_key, data in mesh.vertices(data=True)]
    # closest_index = compas.geometry.closest_point_in_cloud(pt, cloud)[2]
    vertex_tupples = [(v_key, Point(data['x'], data['y'], data['z'])) for v_key, data in mesh.vertices(data=True)]
    vertex_tupples = sorted(vertex_tupples, key=lambda v_tupple: distance_point_point_sqrd(pt, v_tupple[1]))
    closest_vkey = vertex_tupples[0][0]
    return closest_vkey


def get_closest_mesh_normal_to_pt(mesh, pt):
    """
    Finds the closest vertex normal to the point.

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    pt: :class: 'compas.geometry.Point'

    Returns
    ----------
    :class: 'compas.geometry.Vector'
        The closest normal of the mesh.

    """

    closest_vkey = get_closest_mesh_vkey_to_pt(mesh, pt)
    v = mesh.vertex_normal(closest_vkey)
    return Vector(v[0], v[1], v[2])


def get_mesh_vertex_coords_with_attribute(mesh, attr, value):
    """
    Finds the coordinates of all the vertices that have an attribute with key=attr that equals the value.

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    attr: str
    value: anything that can be stored into a dictionary

    Returns
    ----------
    list, :class: 'compas.geometry.Point'
        the closest vertex key
    """

    pts = []
    for vkey, data in mesh.vertices(data=True):
        if data[attr] == value:
            pts.append(Point(*mesh.vertex_coordinates(vkey)))
    return pts


def get_normal_of_path_on_xy_plane(k, point, path, mesh):
    """
    Finds the normal of the curve that lies on the xy plane at the point with index k

    Parameters
    ----------
    k: int, index of the point
    point: :class: 'compas.geometry.Point'
    path: :class: 'compas_slicer.geometry.Path'
    mesh: :class: 'compas.datastructures.Mesh'

    Returns
    ----------
    :class: 'compas.geometry.Vector'
    """

    # find mesh normal is not really needed in the 2D case of planar slicer
    # instead we only need the normal of the curve based on the neighboring pts
    if (0 < k < len(path.points) - 1) or path.is_closed:
        prev_pt = path.points[k - 1]
        next_pt = path.points[(k + 1) % len(path.points)]
        v1 = np.array(normalize_vector(Vector.from_start_end(prev_pt, point)))
        v2 = np.array(normalize_vector(Vector.from_start_end(point, next_pt)))
        v = (v1 + v2) * 0.5
        normal = [-v[1], v[0], v[2]]  # rotate 90 degrees COUNTER-clockwise on the xy plane

    else:
        if k == 0:
            next_pt = path.points[k + 1]
            v = normalize_vector(Vector.from_start_end(point, next_pt))
            normal = [-v[1], v[0], v[2]]  # rotate 90 degrees COUNTER-clockwise on the xy plane
        else:  # k == len(path.points)-1:
            prev_pt = path.points[k - 1]
            v = normalize_vector(Vector.from_start_end(point, prev_pt))
            normal = [v[1], -v[0], v[2]]  # rotate 90 degrees clockwise on the xy plane

    if length_vector(normal) == 0:
        # When the neighboring elements happen to cancel out, then search for the true normal,
        # and project it on the xy plane for consistency
        normal = get_closest_mesh_normal_to_pt(mesh, point)
        normal = [normal[0], normal[1], 0]

    normal = normalize_vector(normal)
    normal = Vector(*list(normal))
    return normal


#######################################
# igl utils

def get_mesh_cotmatrix_igl(mesh, fix_boundaries=True):
    """
    Gets the laplace operator of the mesh

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    fix_boundaries : bool

    Returns
    ----------
    :class: 'scipy.sparse.csr_matrix'
        sparse matrix (dimensions: #V x #V), laplace operator, each row i corresponding to v(i, :)
    """
    import igl
    v, f = mesh.to_vertices_and_faces()
    C = igl.cotmatrix(np.array(v), np.array(f))

    if fix_boundaries:
        # fix boundaries by putting the corresponding columns of the sparse matrix to 0
        C_dense = C.toarray()
        for i, (vkey, data) in enumerate(mesh.vertices(data=True)):
            if data['boundary'] > 0:
                C_dense[i][:] = np.zeros(len(v))
        C = scipy.sparse.csr_matrix(C_dense)
    return C


def get_mesh_cotans_igl(mesh):
    """
    Gets the cotangent entries of the mesh


    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'

    Returns
    ----------
    :class: 'np.array'
        Dimensions: F by 3 list of 1/2*cotangents corresponding angles
    """
    import igl
    v, f = mesh.to_vertices_and_faces()
    return igl.cotmatrix_entries(np.array(v), np.array(f))


#######################################
#  networkx graph

def plot_networkx_graph(G):
    """
    Plots the graph G

    Parameters
    ----------
    G: networkx.Graph
    """

    plt.subplot(121)
    nx.draw(G, with_labels=True, font_weight='bold', node_color=range(len(list(G.nodes()))))
    plt.show()


#######################################
#  dict utils

def point_list_to_dict(pts_list):
    """
    Turns a list of compas.geometry.Point into a dictionary, so that it can be saved to Json. Works identically for
    3D vectors.

    Parameters
    ----------
    pts_list: list, :class:`compas.geometry.Point` / :class:`compas.geometry.Vector`

    Returns
    ----------
    dict: The dictionary of pts in the form { key=index : [x,y,z] }
    """
    data = {}
    for i in range(len(pts_list)):
        data[i] = list(pts_list[i])
    return data


def point_list_from_dict(data):
    """
    Turns a dictionary of pts to a list of Compas.geometry.Point. Works identically for 3D vectors.

    Parameters
    ----------
    dict: The dictionary of pts in the form { key=index : [x,y,z] }

    Returns
    ----------
    2D list,  [[x1, y1, z1], ... , [xn, yn, zn]]
    """
    return [[data[i][0], data[i][1], data[i][2]] for i in data]


#  --- Flattened list of dictionary
def flattened_list_of_dictionary(dictionary):
    """
    Turns the dictionary into a flat list

    Parameters
    ----------
    dictionary: dict

    Returns
    ----------
    list
    """
    flattened_list = []
    for key in dictionary:
        [flattened_list.append(item) for item in dictionary[key]]
    return flattened_list


def get_dict_key_from_value(dictionary, val):
    """
    Return the key of a dictionary that stores the val

    Parameters
    ----------
    dictionary: dict
    val: anything that can be stored in a dictionary
    """

    for key in dictionary:
        value = dictionary[key]
        if val == value:
            return key
    return "key doesn't exist"


def find_next_printpoint(pp_dict, layer_key, path_key, i, j, k):
    """
    Returns the next printpoint from the current printpoint if it exists, otherwise returns None.
    """
    next_ppt = None
    if k < len(pp_dict[layer_key][path_key]) - 1:  # If there are more ppts in the current path, then take the next ppt
        next_ppt = pp_dict[layer_key][path_key][k + 1]
    else:
        if j < len(pp_dict[layer_key]) - 1:  # Otherwise take the next path if there are more paths in the current layer
            next_ppt = pp_dict[layer_key]['path_%d' % (j + 1)][0]
        else:
            if i < len(pp_dict) - 1:  # Otherwise take the next layer if there are more layers in the current slicer
                next_ppt = pp_dict['layer_%d' % (i + 1)]['path_0'][0]
    return next_ppt


def find_previous_printpoint(pp_dict, layer_key, path_key, i, j, k):
    """
    Returns the previous printpoint from the current printpoint if it exists, otherwise returns None.
    """
    prev_ppt = None
    if k > 0:  # If not the first point in a path, take the previous point in the path
        prev_ppt = pp_dict[layer_key][path_key][k - 1]
    else:
        if j > 0:  # Otherwise take the last point of the previous path, if there are more paths in the current layer
            prev_ppt = pp_dict[layer_key]['path_%d' % (j - 1)][-1]
        else:
            if i > 0:  # Otherwise take the last path of the previous layer if there are more layers in the current slicer
                last_path_key = len(pp_dict[layer_key]) - 1
                prev_ppt = pp_dict['layer_%d' % (i - 1)]['path_%d' % (last_path_key)][-1]
    return prev_ppt


#######################################
#  control flow

def interrupt():
    """
    Interrupts the flow of the code while it is running.
    It asks for the user to press a enter to continue or abort.
    """

    value = input("Press enter to continue, Press 1 to abort ")
    print("")
    if isinstance(value, str):
        if value == '1':
            raise ValueError("Aborted")


#######################################
#  load all files with name

def get_all_files_with_name(startswith, endswith, DATA_PATH):
    """
    Finds all the filenames in the DATA_PATH that start and end with the provided strings

    Parameters
    ----------
    startswith: str
    endswith: str
    DATA_PATH: str

    Returns
    ----------
    list, str
        All the filenames
    """

    files = []
    for file in os.listdir(DATA_PATH):
        if file.startswith(startswith) and file.endswith(endswith):
            files.append(file)
    print('')
    logger.info('Reloading : ' + str(files))
    return files


if __name__ == "__main__":
    pass



import compas
import compas_slicer
from compas.geometry import closest_point_on_polyline, distance_point_point, Polyline, Vector, normalize_vector, Point
import logging
from compas_slicer.geometry import Path, PrintPoint
import compas_slicer.utilities as utils
import progressbar
from compas_slicer.parameters import get_param

logger = logging.getLogger('logger')
__all__ = ['VerticalConnectivity']


class VerticalConnectivity:
    """
    VerticalConnectivity finds the vertical relation between paths in a segment.
    It assumes that each path is supported by the path below, and the first path is
    supported by the BaseBoundary.
    This class creates PrintPoints and fills in their information.

    Attributes
    ----------
    paths : list of instances of compas_slicer.geometry.Path
    base_boundary : compas_slicer.print_organization.BaseBoundary
    mesh : compas.geometry.Mesh
    parameters : dict
    """

    def __init__(self, paths, base_boundary, mesh, parameters):
        assert isinstance(paths[0], compas_slicer.geometry.Path)
        assert isinstance(base_boundary, compas_slicer.print_organization.BaseBoundary)
        assert isinstance(mesh, compas.datastructures.Mesh)

        self.paths = paths
        self.base_boundary = base_boundary
        self.mesh = mesh
        self.parameters = parameters
        self.printpoints = {}  # dict with pne list of printpoints per path

    def __repr__(self):
        return "<VerticalConnectivity with %i paths>" % len(self.paths)

    ######################
    # Main
    def compute(self):
        """ Run the segment connectivity process. """
        self.initialize_printpoints()
        self.fill_in_printpoints_information()

        h_smoothing = get_param(self.parameters, key='layer_heights_smoothing', defaults_type='print_organization')
        if h_smoothing[0]:
            logger.info('Layer heights smoothing using horizontal neighbors')
            self.smooth_printpoints_layer_heights(iterations=h_smoothing[1], strength=h_smoothing[2])
        v_smoothing = get_param(self.parameters, key='up_vectors_smoothing', defaults_type='print_organization')
        if v_smoothing[0]:
            logger.info('Up vectors smoothing using horizontal neighbors')
            self.smooth_up_vectors(iterations=v_smoothing[1], strength=v_smoothing[2])

    ######################
    # Utilities
    def initialize_printpoints(self):
        """ Initializes printpoints in a list, without filling in their attributes. """
        for i, path in enumerate(self.paths):
            self.printpoints[i] = [PrintPoint(pt=p, layer_height=None,
                                              mesh_normal=utils.get_closest_mesh_normal_to_pt(self.mesh, p))
                                   for p in path.points]

    def fill_in_printpoints_information(self):
        """ Fills in the attributes of previously initialized printpoints. """
        max_layer_height = get_param(self.parameters, key='max_layer_height', defaults_type='layers')
        min_layer_height = get_param(self.parameters, key='min_layer_height', defaults_type='layers')

        crv_to_check = Path(self.base_boundary.points, True)  # creation of fake path for the lower boundary
        with progressbar.ProgressBar(max_value=len(self.paths)) as bar:
            for i, path in enumerate(self.paths):
                for j, p in enumerate(path.points):
                    cp = closest_point_on_polyline(p, Polyline(crv_to_check.points))
                    d = distance_point_point(cp, p)

                    self.printpoints[i][j].closest_support_pt = Point(*cp)
                    self.printpoints[i][j].distance_to_support = d
                    self.printpoints[i][j].layer_height = max(min(d, max_layer_height), min_layer_height)
                    self.printpoints[i][j].up_vector = Vector(*normalize_vector(Vector.from_start_end(cp, p)))
                    self.printpoints[i][j].frame = self.printpoints[i][j].get_frame()

                bar.update(i)
                crv_to_check = path

    def smooth_printpoints_layer_heights(self, iterations, strength):
        """ Smooth printpoints heights based on neighboring. """
        for i, path in enumerate(self.paths):
            self.smooth_path_printpoint_attribute(i, iterations, strength, get_attr_value=lambda pp: pp.layer_height,
                                                  set_attr_value=set_printpoint_height)

    def smooth_up_vectors(self, iterations, strength):
        """ Smooth printpoints up_vectors based on neighboring. """
        for i, path in enumerate(self.paths):
            self.smooth_path_printpoint_attribute(i, iterations, strength, get_attr_value=lambda pp: pp.up_vector,
                                                  set_attr_value=set_printpoint_up_vec)

    def smooth_path_printpoint_attribute(self, path_index, iterations, strength, get_attr_value, set_attr_value):

        """General description.
        Parameters
        ----------
        path_index: int
        iterations: int
        strength: float
        get_attr_value: function that returns an attribute of a printpoint, get_attr_value(pp)
        set_attr_value: function that sets an attribute of a printpoint, set_attr_value(pp, new_value)
        """
        for _ in range(iterations):
            attributes = [get_attr_value(pp) for pp in self.printpoints[path_index]]

            if self.paths[path_index].is_closed:
                for j, v in enumerate(attributes):
                    mid = (attributes[j - 1] + attributes[(j + 1) % len(attributes)]) / 2
                    new_value = mid * strength + v * (1 - strength)
                    set_attr_value(self.printpoints[path_index][j], new_value)
            else:
                for j, v in enumerate(attributes):
                    if 0 < j < len(attributes) - 1:
                        mid = (attributes[j - 1] + attributes[(j + 1)]) / 2
                        new_value = mid * strength + v * (1 - strength)
                        set_attr_value(self.printpoints[path_index][j], new_value)


def set_printpoint_up_vec(pp, v):
    """ Set printpoint.up_vector = v. """
    pp.up_vector = v


def set_printpoint_height(pp, h):
    """ Set printpoint.layer_height = h. """
    pp.layer_height = h


if __name__ == "__main__":
    pass



import math
from compas.geometry import Vector, angle_vectors
from compas_slicer.utilities import get_closest_mesh_normal_to_pt
import logging

logger = logging.getLogger('logger')

__all__ = ['set_linear_velocity']


def set_linear_velocity(print_organizer,
                        velocity_type,
                        v=25.0,
                        per_layer_velocities=None,
                        angle_range=(None, None),
                        speed_range=(None, None)):
    """ Sets the linear velocity parameter of the printpoints depending on the selected type.

    Parameters
    ----------
    print_organizer: :class:`compas_slicer.print_organization.BasePrintOrganizer`
    velocity_type: str
        Determines how to add linear velocity to the printpoints.

        'constant':              one value used for all printpoints
        'per_layer':             different values used for every layer
        'by_layer_height':       set velocity in accordance to layer height (beta, to be improved)
        'by_overhang':           set velocity in accordance to the overhang (beta, to be improved)
    v: float
        Velocity value (in mm/s) to set for printpoints. Defaults to 25 mm/s.
    per_layer_velocities: list of floats
        If setting velocity per layer, provide a list of floats with equal length to the number of layers.
    angle_range: tuple
        Two angles (in degrees) that give the minimum and maximum value to map the speed_range to.
        For example, (10, 50) will assign the maximum speed to overhang angles of 10 degrees
        or less, and the minimum speed to points with an overhang angle of 50 degrees and above.
    speed_range: tuple
        xxx
    """
    logger.info("Setting linear velocity with type : " + str(velocity_type))

    if not (velocity_type == "constant"
            or velocity_type == "per_layer"
            or velocity_type == "by_layer_height"
            or velocity_type == "by_overhang"):
        raise ValueError("Velocity method doesn't exist")

    pp_dict = print_organizer.printpoints_dict

    for i, layer_key in enumerate(pp_dict):
        for path_key in pp_dict[layer_key]:
            path_printpoints = pp_dict[layer_key][path_key]
            for printpoint in path_printpoints:

                if velocity_type == "constant":
                    printpoint.velocity = v

                elif velocity_type == "per_layer":
                    assert per_layer_velocities, "You need to provide one velocity value per layer"
                    assert len(per_layer_velocities) == print_organizer.number_of_layers, \
                        'Wrong number of velocity values. You need to provide one velocity value per layer, ' \
                        'on the "per_layer_velocities" list.'
                    printpoint.velocity = per_layer_velocities[i]

                # ----- TODO: these two methods are trying to do the same thing. They should be merged!
                elif velocity_type == "by_layer_height":
                    printpoint.velocity = calculate_linear_velocity_based_on_layer_height(printpoint)
                elif velocity_type == "by_overhang":
                    mesh = print_organizer.slicer.mesh
                    printpoint.velocity = calculate_linear_velocity_based_on_overhang(printpoint, mesh,
                                                                                      angle_range, speed_range)


# --------  Nozzle linear velocity based on layer height

# TODO: These parameters should NOT be hardcoded! Instead they should be passed as a dictionary of parameters.
motor_omega = 2 * math.pi  # 1 revolution / sec = 2*pi rad/sec
motor_r = 4.0  # 4.25 #mm
motor_linear_speed = motor_omega * motor_r
D_filament = 2.75  # mm
filament_area = math.pi * (D_filament / 2.0) ** 2  # pi*r^2
multiplier = 0.25  # arbitrary value! You might have to change this


def calculate_linear_velocity_based_on_layer_height(printpoint):
    """
    Calculates the linear velocity during the print of a printpoint, so that the volume of extrusion matches the
    volume of the desired path cross-section
    path_area * robot_linear_speed = filament_area * motor_linear_speed

    Parameters
    ----------
    printpoint: :class: 'compas_slicer.geometry.PrintPoint'
    """
    layer_width = max(printpoint.layer_height, 0.4)
    path_area = layer_width * printpoint.layer_height
    linear_speed = (filament_area * motor_linear_speed) / path_area
    return linear_speed * multiplier


# --------  Nozzle linear velocity based on overhang
def calculate_linear_velocity_based_on_overhang(printpoint, mesh, angle_range, speed_range):
    """
    Calculates the linear velocity during the print of a printpoint, so that it is slower on large overhangs.

    Parameters
    ----------
    printpoint: :class: 'compas_slicer.geometry.PrintPoint'
    mesh: :class: 'compas.datastructures.Mesh'
    angle_range: tupple
    speed_range: tupple

    """
    # get mesh normal and calculate overhang angle
    mesh_normal = get_closest_mesh_normal_to_pt(mesh, printpoint.pt)
    vect_angle = angle_vectors(mesh_normal, Vector(0.0, 0.0, 1.0), deg=True)
    overhang_angle = abs(90 - vect_angle)

    # remap values
    min_speed, max_speed = speed_range
    min_angle, max_angle = angle_range
    old_range = max_angle - min_angle
    new_range = max_speed - min_speed

    if overhang_angle <= min_angle:
        return max_speed
    elif min_angle < overhang_angle <= max_angle:
        velocity = (((overhang_angle - min_angle) * new_range) / old_range) + min_speed
        return new_range - (velocity - new_range)
    else:  # overhang_angle > max_angle:
        return min_speed


if __name__ == "__main__":
    pass



"""
********************************************************************************
utilities
********************************************************************************

.. currentmodule:: compas_slicer.utilities


utils
=========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    save_to_json
    load_from_json
    flattened_list_of_dictionary
    interrupt
    point_list_to_dict
    get_closest_mesh_normal_to_pt
    get_closest_pt_index
    get_closest_pt
    plot_networkx_graph
    get_mesh_vertex_coords_with_attribute
    get_dict_key_from_value
    get_closest_mesh_normal_to_pt
    smooth_vectors
    get_normal_of_path_on_xy_plane

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .utils import *  # noqa: F401 E402 F403
from .terminal_command import *  # noqa: F401 E402 F403
from .attributes_transfer import *  # noqa: F401 E402 F403

__all__ = [name for name in dir() if not name.startswith('_')]



import networkx as nx
from compas.geometry import distance_point_point_sqrd
import compas_slicer.utilities as utils
import logging
import copy
from compas_slicer.pre_processing.preprocessing_utils import get_existing_cut_indices, \
    get_existing_boundary_indices
from abc import abstractmethod

logger = logging.getLogger('logger')

__all__ = ['MeshDirectedGraph',
           'SegmentsDirectedGraph']


#################################
#  DirectedGraph

class DirectedGraph(object):
    """
    Base class for topological sorting of prints that consist of several parts that lie on each other.
    For example the graph A->B->C would represent a print that consists of three parts; A, B, C
    where A lies on the build platform, B lies on A, and C lies on B.
    This graph cannot have cycles; cycles would represent an unfeasible print.
    """

    def __init__(self):
        logger.info('Topological sorting')

        self.G = nx.DiGraph()
        self.create_graph_nodes()
        self.root_indices = self.find_roots()
        logger.info('Graph roots : ' + str(self.root_indices))
        assert len(self.root_indices) > 0, "No root nodes where found. At least one root node is needed."
        self.end_indices = self.find_ends()
        logger.info('Graph ends : ' + str(self.end_indices))
        assert len(self.end_indices) > 0, "No end nodes where found. At least one end node is needed."

        self.create_directed_graph_edges(copy.deepcopy(self.root_indices))
        logger.info('Nodes : ' + str(self.G.nodes(data=True)))
        logger.info('Edges : ' + str(self.G.edges(data=True)))

        self.N = len(list(self.G.nodes()))
        self.adj_list = self.get_adjacency_list()  # Nested list where adj_list[i] is a list of all the neighbors
        # of the i-th component
        self.check_that_all_nodes_found_their_connectivity()
        logger.info('Adjacency list : ' + str(self.adj_list))
        self.in_degree = self.get_in_degree()  # Nested list where adj_list[i] is a list of all the edges pointing
        # to the i-th node.
        self.all_orders = []

    def __repr__(self):
        return "<DirectedGraph with %i nodes>" % len(list(self.G.nodes()))

    # ------------------------------------ Methods to be implemented by inheriting classes
    @abstractmethod
    def find_roots(self):
        """ Roots are vertical_layers_print_data that lie on the build platform. Like that they can be print first. """
        pass

    @abstractmethod
    def find_ends(self):
        """ Ends are vertical_layers_print_data that belong to exclusively one segment. Like that they can be print last. """
        pass

    @abstractmethod
    def create_graph_nodes(self):
        """ Add the nodes to the graph with their attributes. """
        pass

    @abstractmethod
    def get_children_of_node(self, root):
        """ Find all the vertical_layers_print_data that lie on the current root segment. """
        pass

    # ------------------------------------ Creation of graph connectivity between different nodes
    def create_directed_graph_edges(self, root_indices):
        """ Create the connectivity of the directed graph using breadth-first search graph traversal. """
        passed_nodes = []
        queue = root_indices

        while len(queue) > 0:
            queue = self.sort_queue_with_end_targets_last(queue)
            current_node = queue[0]
            queue.remove(current_node)
            passed_nodes.append(current_node)
            children, cut_ids = self.get_children_of_node(current_node)
            [self.G.add_edge(current_node, child_key, cut=common_cuts) for child_key, common_cuts in
             zip(children, cut_ids)]
            for child_key in children:
                assert child_key not in passed_nodes, 'Error, cyclic directed graph.'
            [queue.append(child_key) for child_key in children if child_key not in queue]

    def check_that_all_nodes_found_their_connectivity(self):
        """ Assert that there is no island, i.e. no node or groups of nodes that are not connected to the base. """
        good_nodes = [r for r in self.root_indices]
        for children_list in self.adj_list:
            [good_nodes.append(child) for child in children_list if child not in good_nodes]
        assert len(good_nodes) == self.N, 'There are floating vertical_layers_print_data on directed graph. Investigate the process of \
                                          the creation of the graph. '

    def sort_queue_with_end_targets_last(self, queue):
        """ Sorts the queue so that the vertical_layers_print_data that have an end target are always at the end. """
        queue_copy = copy.deepcopy(queue)
        for index in queue:
            if index in self.end_indices:
                queue_copy.remove(index)  # remove it from its current position
                queue_copy.append(index)  # append it last
        return queue_copy

    # ------------------------------------ Find all topological orders
    def get_adjacency_list(self):
        """ Returns adjacency list. Nested list where adj_list[i] is a list of all the neighbors of the ith component"""
        adj_list = [[] for _ in range(self.N)]  # adjacency list , size = len(Nodes), stores nodes' neighbors
        for i, adjacent_to_node in self.G.adjacency():
            [adj_list[i].append(key) for key in adjacent_to_node]
        return adj_list

    def get_in_degree(self):
        """ Returns in_degree list. Nested list where adj_list[i] is a list of all the edges pointing to the node."""
        in_degree = [0] * self.N  # in_degree,  size = len(Nodes) , stores in-degree of a node
        for key_degree_tuple in self.G.in_degree:
            key = key_degree_tuple[0]
            degree = key_degree_tuple[1]
            in_degree[key] = degree
        return in_degree

    def get_all_topological_orders(self):
        """
        Finds  all topological orders from source to sink.
        Returns
        ----------
        list of lists of integers. Each list represents the indices of one topological order.
        """
        self.all_orders = []  # make sure list is empty
        discovered = [False] * self.N
        path = []  # list to store the topological order
        self.get_orders(path, discovered)
        logger.info('Found %d possible orders' % len(self.all_orders))
        return self.all_orders

    def get_orders(self, path, discovered):
        """
        Finds all topological orders from source to sink.
        Sorting algorithm taken from https://www.techiedelight.com/find-all-possible-topological-orderings-of-dag/
        """
        for v in range(self.N):  # for every node
            # proceed only if in-degree of current node is 0 and current node is not processed yet
            if self.in_degree[v] == 0 and not discovered[v]:

                # for every adjacent vertex u of v, reduce in-degree of u by 1
                for u in self.adj_list[v]:
                    self.in_degree[u] = self.in_degree[u] - 1

                # include current node in the path and mark it as discovered
                path.append(v)
                discovered[v] = True

                # recur
                self.get_orders(path, discovered)

                # backtrack: reset in-degree information for the current node
                for u in self.adj_list[v]:
                    self.in_degree[u] = self.in_degree[u] + 1

                # backtrack: remove current node from the path and mark it as undiscovered
                path.pop()
                discovered[v] = False

        # print the topological order if all vertices are included in the path
        if len(path) == self.N:
            self.all_orders.append(copy.deepcopy(path))

    def get_parents_of_node(self, node_index):
        """ Returns the parents of node with i = node_index. """
        return [j for j, adj in enumerate(self.adj_list) if node_index in adj]


#################################
#  --- Meshes DirectedGraph


class MeshDirectedGraph(DirectedGraph):
    """ The MeshDirectedGraph is used for topological sorting of multiple meshes that have been
    generated as a result of region split over the saddle points of the mesh scalar function """

    def __init__(self, all_meshes, DATA_PATH):
        self.all_meshes = all_meshes
        self.DATA_PATH = DATA_PATH
        self.OUTPUT_PATH = utils.get_output_directory(DATA_PATH)
        DirectedGraph.__init__(self)

    def find_roots(self):
        """ Roots are vertical_layers_print_data that lie on the build platform. Like that they can be print first. """
        roots = []
        for i, mesh in enumerate(self.all_meshes):
            for vkey, data in mesh.vertices(data=True):
                if i not in roots:
                    if data['boundary'] == 1:
                        roots.append(i)
        return roots

    def find_ends(self):
        """ Ends are vertical_layers_print_data that belong to exclusively one segment. Like that they can be print last. """
        ends = []
        for i, mesh in enumerate(self.all_meshes):
            for vkey, data in mesh.vertices(data=True):
                if i not in ends:
                    if data['boundary'] == 2:
                        ends.append(i)
        return ends

    def create_graph_nodes(self):
        """ Add each of the split meshes to the graph as nodes. Cuts and boundaries are stored as attributes. """
        for i, m in enumerate(self.all_meshes):
            self.G.add_node(i, cuts=get_existing_cut_indices(m),
                            boundaries=get_existing_boundary_indices(m))

    def get_children_of_node(self, root):
        """
        Find all the nodes that lie on the current root.

        Parameters
        ----------
        root: int, index of root node

        Returns
        ----------
        2 lists [child1, child2, ...], [[common cuts 1], [common cuts 2] ...]
        """
        children = []
        cut_ids = []
        parent_data = self.G.nodes(data=True)[root]

        for key, data in self.G.nodes(data=True):
            common_cuts = list(set(parent_data['cuts']).intersection(data['cuts']))

            if key != root and len(common_cuts) > 0 \
                    and (key, root) not in self.G.edges() \
                    and (root, key) not in self.G.edges():

                if is_true_mesh_adjacency(self.all_meshes, key, root):
                    if not len(common_cuts) == 1:  # if all cuts worked, this should be 1. But life is not perfect.
                        logger.error('More than one common cuts between two pieces in the following split \
                        meshes. ' 'Root : %d, child : %d' % (root, key) + ' . Common cuts : ' + str(common_cuts) +
                                     'Probably some cut did not separate components')
                    children.append(key)
                    cut_ids.append(common_cuts)

        # --- debugging output
        # self.all_meshes[root].to_obj(self.OUTPUT_PATH + '/root.obj')
        # for child in children:
        #     self.all_meshes[child].to_obj(self.OUTPUT_PATH + '/child_%d.obj' % child)
        # for cuts_id in cut_ids:
        #     for common_cut in cuts_id:
        #         pts = utils.get_mesh_vertex_coords_with_attribute(self.all_meshes[root], 'cut', common_cut)
        #         utils.save_to_json(utils.point_list_to_dict(pts), self.OUTPUT_PATH, 'cut_%d.json' % common_cut)
        # print('root : ', root)
        # print('children : ', children)
        # print('cut_ids : ', cut_ids)
        # utils.interrupt()

        return children, cut_ids


#################################
#  --- Segments DirectedGraph

class SegmentsDirectedGraph(DirectedGraph):
    """ The SegmentsDirectedGraph is used for topological sorting of multiple vertical_layers_print_data in one mesh"""

    def __init__(self, mesh, segments, max_d_threshold, DATA_PATH):
        self.mesh = mesh
        self.segments = segments
        self.max_d_threshold = max_d_threshold
        self.DATA_PATH = DATA_PATH
        self.OUTPUT_PATH = utils.get_output_directory(DATA_PATH)
        DirectedGraph.__init__(self)

    def find_roots(self):
        """ Roots are vertical_layers_print_data that lie on the build platform. Like that they can be print first. """
        boundary_pts = utils.get_mesh_vertex_coords_with_attribute(self.mesh, 'boundary', 1)
        root_segments = []
        for i, segment in enumerate(self.segments):
            first_curve_pts = segment.paths[0].points
            if are_neighboring_point_clouds(boundary_pts, first_curve_pts, self.max_d_threshold):
                root_segments.append(i)
        return root_segments

    def find_ends(self):
        """ Ends are vertical_layers_print_data that belong to exclusively one segment. Like that they can be print last. """
        boundary_pts = utils.get_mesh_vertex_coords_with_attribute(self.mesh, 'boundary', 2)
        end_segments = []
        for i, segment in enumerate(self.segments):
            last_curve_pts = segment.paths[-1].points
            if are_neighboring_point_clouds(boundary_pts, last_curve_pts, self.max_d_threshold):
                end_segments.append(i)
        return end_segments

    def create_graph_nodes(self):
        """ Add each segment to to the graph as a node. """
        for i, segment in enumerate(self.segments):
            self.G.add_node(i)

    def get_children_of_node(self, root):
        """ Find all the nodes that lie on the current root. """
        children = []
        root_segment = self.segments[root]
        root_last_crv_pts = root_segment.paths[-1].points

        for i, segment in enumerate(self.segments):
            if i != root:
                segment_first_curve_pts = segment.paths[0].points
                if are_neighboring_point_clouds(root_last_crv_pts, segment_first_curve_pts, self.max_d_threshold):
                    children.append(i)
        return children, [None for _ in children]  # None because this graph doesn't have cut ids


#################################
# --- helpers

def are_neighboring_point_clouds(pts1, pts2, threshold):
    """
    Returns True if 3 or more points of the point clouds are closer than the threshold. False otherwise.

    Parameters
    ----------
    pts1: list, :class: 'compas.geometry.Point'
    pts2: list, :class: 'compas.geometry.Point'
    threshold: float
    """
    count = 0
    for pt in pts1:
        if distance_point_point_sqrd(pt, utils.get_closest_pt(pt, pts2)) < threshold:
            count += 1
            if count > 2:
                return True
    return False


def is_true_mesh_adjacency(all_meshes, key1, key2):
    """
    Returns True if the two meshes share 3 or more vertices. False otherwise.

    Parameters
    ----------
    all_meshes: list, :class: 'compas.datastructures.Mesh'
    key1: int, index of mesh1
    key2: int, index of mesh2
    """
    count = 0
    mesh1 = all_meshes[key1]
    mesh2 = all_meshes[key2]
    pts_mesh2 = [mesh2.vertex_coordinates(vkey) for vkey, data in mesh2.vertices(data=True)
                 if (data['cut'] > 0 or data['boundary'] > 0)]
    for vkey, data in mesh1.vertices(data=True):
        if data['cut'] > 0 or data['boundary'] > 0:
            pt = mesh1.vertex_coordinates(vkey)
            ci = utils.get_closest_pt_index(pt, pts_mesh2)
            if distance_point_point_sqrd(pt, pts_mesh2[ci]) < 0.00001:
                count += 1
                if count == 3:
                    return True
    return False


if __name__ == '__main__':
    pass



import numpy as np
from compas_slicer.slicers import BaseSlicer
import logging
from compas_slicer.slicers.slice_utilities import ScalarFieldContours
import progressbar
from compas_slicer.geometry import VerticalLayersManager
from compas_slicer.parameters import get_param

logger = logging.getLogger('logger')

__all__ = ['ScalarFieldSlicer']


class ScalarFieldSlicer(BaseSlicer):
    """
    Generates the isocontours of a scalar field defined on the mesh vertices.

    Attributes
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
        Input mesh, it must be a triangular mesh (i.e. no quads or n-gons allowed)
        Note that the topology of the mesh matters, irregular tesselation can lead to undesired results.
        We recommend to 1)re-topologize, 2) triangulate, and 3) weld your mesh in advance.
    scalar_field: list, Vx1 (one float per vertex that represents the scalar field)
    no_of_isocurves: int, how many isocontours to be generated
    """

    def __init__(self, mesh, scalar_field, no_of_isocurves, parameters=None):
        logger.info('ScalarFieldSlicer')
        BaseSlicer.__init__(self, mesh)

        self.no_of_isocurves = no_of_isocurves
        self.scalar_field = list(np.array(scalar_field) - np.max(np.array(scalar_field)))
        self.parameters = parameters if parameters else {}

        mesh.update_default_vertex_attributes({'scalar_field': 0})

    def generate_paths(self):
        """ Generates isocontours. """
        start_domain, end_domain = min(self.scalar_field), max(self.scalar_field)
        step = (end_domain - start_domain) / (self.no_of_isocurves + 1)

        max_dist = get_param(self.parameters, key='vertical_layers_max_centroid_dist', defaults_type='layers')
        vertical_layers_manager = VerticalLayersManager(max_dist)

        # create paths + layers
        with progressbar.ProgressBar(max_value=self.no_of_isocurves) as bar:
            for i in range(1, self.no_of_isocurves + 1):
                for vkey, data in self.mesh.vertices(data=True):
                    data['scalar_field'] = self.scalar_field[vkey] + i * step

                contours = ScalarFieldContours(self.mesh)
                contours.compute()
                contours.add_to_vertical_layers_manager(vertical_layers_manager)

                bar.update(i)  # advance progress bar

        self.layers = vertical_layers_manager.layers


if __name__ == "__main__":
    pass



import compas_slicer.utilities as utils
from compas.geometry import Point, distance_point_point_sqrd

__all__ = ['create_mesh_boundary_attributes',
           'get_existing_cut_indices',
           'get_existing_boundary_indices',
           'get_vertices_that_belong_to_cuts',
           'save_vertex_attributes',
           'restore_mesh_attributes',
           'replace_mesh_vertex_attribute']


def create_mesh_boundary_attributes(mesh, low_boundary_vs, high_boundary_vs):
    """
    Creates a default vertex attribute data['boundary']=0. Then it gives the value 1 to the vertices that belong
    to the lower boundary, and the value 2 to the vertices that belong to the higher boundary.
    """
    mesh.update_default_vertex_attributes({'boundary': 0})
    for vkey, data in mesh.vertices(data=True):
        if vkey in low_boundary_vs:
            data['boundary'] = 1
        elif vkey in high_boundary_vs:
            data['boundary'] = 2


###############################################
# --- Mesh existing attributes on vertices

def get_existing_cut_indices(mesh):
    """
    Returns
    ----------
        list, int.
        The cut indices (data['cut']>0) that exist on the mesh vertices.
    """
    cut_indices = []
    for vkey, data in mesh.vertices(data=True):
        if data['cut'] > 0:
            if data['cut'] not in cut_indices:
                cut_indices.append(data['cut'])
    cut_indices = sorted(cut_indices)
    return cut_indices


def get_existing_boundary_indices(mesh):
    """
    Returns
    ----------
        list, int.
        The boundary indices (data['boundary']>0) that exist on the mesh vertices.
    """
    indices = []
    for vkey, data in mesh.vertices(data=True):
        if data['boundary'] > 0:
            if data['boundary'] not in indices:
                indices.append(data['boundary'])
    boundary_indices = sorted(indices)
    return boundary_indices


def get_vertices_that_belong_to_cuts(mesh, cut_indices):
    """
    Returns
    ----------
        dict, key: int, the index of each cut
              value: list, the points that belong to this cut
    """
    cuts_dict = {i: [] for i in cut_indices}

    for vkey, data in mesh.vertices(data=True):
        if data['cut'] > 0:
            cut_index = data['cut']
            cuts_dict[cut_index].append(mesh.vertex_coordinates(vkey))

    for cut_index in cuts_dict:
        cuts_dict[cut_index] = utils.point_list_to_dict(cuts_dict[cut_index])

    return cuts_dict


###############################################
# --- Save and restore attributes

def save_vertex_attributes(mesh):
    """
    Saves the boundary and cut attributes that are on the mesh on a dictionary.
    """
    v_attributes_dict = {'boundary_1': [], 'boundary_2': [], 'cut': {}}

    cut_indices = []
    for vkey, data in mesh.vertices(data=True):
        cut_index = data['cut']
        if cut_index not in cut_indices:
            cut_indices.append(cut_index)
    cut_indices = sorted(cut_indices)

    for cut_index in cut_indices:
        v_attributes_dict['cut'][cut_index] = []

    for vkey, data in mesh.vertices(data=True):
        if data['boundary'] == 1:
            v_coords = mesh.vertex_coordinates(vkey)
            pt = Point(x=v_coords[0], y=v_coords[1], z=v_coords[2])
            v_attributes_dict['boundary_1'].append(pt)
        elif data['boundary'] == 2:
            v_coords = mesh.vertex_coordinates(vkey)
            pt = Point(x=v_coords[0], y=v_coords[1], z=v_coords[2])
            v_attributes_dict['boundary_2'].append(pt)
        if data['cut'] > 0:
            cut_index = data['cut']
            v_coords = mesh.vertex_coordinates(vkey)
            pt = Point(x=v_coords[0], y=v_coords[1], z=v_coords[2])
            v_attributes_dict['cut'][cut_index].append(pt)
    return v_attributes_dict


def restore_mesh_attributes(mesh, v_attributes_dict):
    """
    Restores the cut and boundary attributes on the mesh vertices from the dictionary of the previously saved attributes
    """
    mesh.update_default_vertex_attributes({'boundary': 0})
    mesh.update_default_vertex_attributes({'cut': 0})

    D_THRESHOLD = 0.01

    welded_mesh_vertices = []
    indices_to_vkeys = {}
    for i, vkey in enumerate(mesh.vertices()):
        v_coords = mesh.vertex_coordinates(vkey)
        pt = Point(x=v_coords[0], y=v_coords[1], z=v_coords[2])
        welded_mesh_vertices.append(pt)
        indices_to_vkeys[i] = vkey

    for v_coords in v_attributes_dict['boundary_1']:
        closest_index = utils.get_closest_pt_index(pt=v_coords, pts=welded_mesh_vertices)
        c_vkey = indices_to_vkeys[closest_index]
        if distance_point_point_sqrd(v_coords, mesh.vertex_coordinates(c_vkey)) < D_THRESHOLD:
            mesh.vertex_attribute(c_vkey, 'boundary', value=1)

    for v_coords in v_attributes_dict['boundary_2']:
        closest_index = utils.get_closest_pt_index(pt=v_coords, pts=welded_mesh_vertices)
        c_vkey = indices_to_vkeys[closest_index]
        if distance_point_point_sqrd(v_coords, mesh.vertex_coordinates(c_vkey)) < D_THRESHOLD:
            mesh.vertex_attribute(c_vkey, 'boundary', value=2)

    for cut_index in v_attributes_dict['cut']:
        for v_coords in v_attributes_dict['cut'][cut_index]:
            closest_index = utils.get_closest_pt_index(pt=v_coords, pts=welded_mesh_vertices)
            c_vkey = indices_to_vkeys[closest_index]
            if distance_point_point_sqrd(v_coords, mesh.vertex_coordinates(c_vkey)) < D_THRESHOLD:
                mesh.vertex_attribute(c_vkey, 'cut', value=int(cut_index))


def replace_mesh_vertex_attribute(mesh, old_attr, old_val, new_attr, new_val):
    """
    Replaces one vertex attribute with a new one. For all the vertices where data[old_attr]=old_val, then the
    old_val is replaced with 0, and data[new_attr]=new_val.
    """
    for vkey, data in mesh.vertices(data=True):
        if data[old_attr] == old_val:
            mesh.vertex_attribute(vkey, old_attr, 0)
            mesh.vertex_attribute(vkey, new_attr, new_val)


if __name__ == "__main__":
    pass



"""
********************************************************************************
print_organization
********************************************************************************

.. currentmodule:: compas_slicer.print_organization


Classes
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    BasePrintOrganizer


BasePrintOrganizer
------------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    InterpolationPrintOrganizer


Functions
=========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    set_extruder_toggle
    override_extruder_toggle
    set_linear_velocity
    set_blend_radius
    add_safety_printpoints
    set_wait_time
    override_wait_time

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from .base_print_organizer import *  # noqa: F401 E402 F403
from .planar_print_organizer import *  # noqa: F401 E402 F403
from .interpolation_print_organizer import *  # noqa: F401 E402 F403
from .scalar_field_print_organizer import *  # noqa: F401 E402 F403

from .curved_print_organization import *  # noqa: F401 E402 F403
from .print_organization_utilities import *  # noqa: F401 E402 F403

__all__ = [name for name in dir() if not name.startswith('_')]



"""
********************************************************************************
slicers
********************************************************************************

.. currentmodule:: compas_slicer.slicers


Classes
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    BaseSlicer


BaseSlicer
----------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    PlanarSlicer
    InterpolationSlicer
"""


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .base_slicer import *  # noqa: F401 E402 F403
from .planar_slicer import *  # noqa: F401 E402 F403
from .interpolation_slicer import *  # noqa: F401 E402 F403
from .planar_slicing import *  # noqa: F401 E402 F403
from .scalar_field_slicer import *  # noqa: F401 E402 F403
from .uv_slicer import *  # noqa: F401 E402 F403


__all__ = [name for name in dir() if not name.startswith('_')]



import networkx as nx

__all__ = ['create_graph_from_mesh_edges',
           'sort_graph_connected_components',
           'create_graph_from_mesh_vkeys']


def create_graph_from_mesh_edges(mesh, intersection_data, edge_to_index):
    """
    Creates a graph with one node for every intersected edge.
    The connectivity of nodes (i.e. edges between them) is based on their neighboring on the mesh.

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    intersection_data: dict, (ui,vi) : {'pt':[xi,yi,zi], 'attrs':{...}}
        The keys-tuples are all the edges which have a zero crossing.
    edge_to_index: dict
        key: tuple (int, int) edge
        value: int, index of the intersection point

    Returns
    ----------
    G: :class: 'networkx.Graph'. Node_indices > intersection indices. Node_attrs > their parent_edges (i.e. the keys of
        the intersection_data dict)
    """

    # create graph
    G = nx.Graph()
    for i, edge in enumerate(intersection_data):
        G.add_node(i, mesh_edge=edge)  # node, attribute

    for node_index, data in G.nodes(data=True):
        mesh_edge = data['mesh_edge']

        # find current neighboring edges
        current_edge_connections = []
        for f in mesh.edge_faces(u=mesh_edge[0], v=mesh_edge[1]):
            if f is not None:
                face_edges = mesh.face_halfedges(f)
                for e in face_edges:
                    if (e != mesh_edge and tuple(reversed(e)) != mesh_edge) \
                            and (e in intersection_data or tuple(reversed(e)) in intersection_data):
                        current_edge_connections.append(e)

        for e in current_edge_connections:
            # find other_node_index
            other_node_index = edge_to_index[e] if e in edge_to_index \
                else edge_to_index[tuple(reversed(e))]
            # add edges to the graph (only if the edge doesn't exist already)
            if not G.has_edge(node_index, other_node_index) and not G.has_edge(other_node_index, node_index):
                G.add_edge(node_index, other_node_index)

    return G


def create_graph_from_mesh_vkeys(mesh, v_keys):
    """
    Creates a graph with one node for every vertex, and edges between neighboring vertices.

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    v_keys: list int
        The vertex keys

    Returns
    ----------
    G: :class: 'networkx.Graph'
    """
    G = nx.Graph()
    [G.add_node(v) for v in v_keys]
    for v in v_keys:
        v_neighbors = mesh.vertex_neighbors(v)
        for other_v in v_neighbors:
            if other_v != v and other_v in v_keys:
                G.add_edge(v, other_v)
    return G


def sort_graph_connected_components(G):
    """
    For every connected component of the graph G:
    1) It finds a start node. For open paths it is on one of its ends, for closed paths it can be any of its points.
    2) It sorts all nodes starting from the start using depth first graph traversal.
    3) Using this sorted order, it sorts the intersected edges and zero-crossing points that are linked to the graph G.

    Parameters
    ----------
    G: :class: 'networkx.Graph', Node_indices > intersection point indices. Node_attrs > parent_edges

    Returns
    ----------
    sorted_point_clusters: dict
        key: int, The index of the connected component
        value: list, :class: 'compas.geometry.Point', The sorted zero-crossing points.
    sorted_edge_clusters: dict
        key: int, The index of the connected component.
        value: list, tuple (int, int), The sorted intersected edges.
    """

    sorted_indices_dict = {}

    for j, cp in enumerate(nx.connected_components(G)):
        sorted_node_indices = []

        # (1) find start_node index
        start_node = None
        for node in cp:
            if not start_node:
                start_node = node
            if (len(list(nx.neighbors(G, node)))) == 1:
                start_node = node
                break

        # (2) sort nodes_indices with depth first graph traversal from start node
        for edge_of_node_indices in nx.dfs_edges(G, start_node):
            # edge_of_node_indices: pair of neighboring graph nodes
            node_index_1 = edge_of_node_indices[0]
            node_index_2 = edge_of_node_indices[1]
            if node_index_1 not in sorted_node_indices:
                sorted_node_indices.append(node_index_1)
            if node_index_2 not in sorted_node_indices:
                sorted_node_indices.append(node_index_2)

        assert len(sorted_node_indices) == len(cp), 'Attention. len(sorted_node_indices) != len(G.nodes())'

        sorted_indices_dict[j] = sorted_node_indices
    return sorted_indices_dict



import compas_slicer
from compas_slicer.slicers import BaseSlicer
from compas.geometry import Vector, Plane, Point
import logging

logger = logging.getLogger('logger')

__all__ = ['PlanarSlicer']


class PlanarSlicer(BaseSlicer):
    """
    Generates planar contours on a mesh that are parallel to the xy plane.

    Attributes
    ----------
    mesh: :class:`compas.datastructures.Mesh`
        Input mesh, it must be a triangular mesh (i.e. no quads or n-gons allowed).
    slicer_type: str
        String representing which slicing method to use.
        options: 'default', 'cgal', 'meshcut'
    layer_height: float
        Distance between layers (slices).
    """
    def __init__(self, mesh, slicer_type="default", layer_height=2.0):
        logger.info('PlanarSlicer')
        BaseSlicer.__init__(self, mesh)

        self.layer_height = layer_height
        self.slicer_type = slicer_type

    def __repr__(self):
        return "<PlanarSlicer with %d layers and layer_height : %.2f mm>" % \
               (len(self.layers), self.layer_height)

    def generate_paths(self):
        """Generates the planar slicing paths."""
        z = [self.mesh.vertex_attribute(key, 'z') for key in self.mesh.vertices()]
        min_z, max_z = min(z), max(z)
        d = abs(min_z - max_z)
        no_of_layers = int(d / self.layer_height) + 1
        normal = Vector(0, 0, 1)
        planes = [Plane(Point(0, 0, min_z + i * self.layer_height), normal) for i in range(no_of_layers)]
        planes.pop(0)  # remove planes that are on the print platform

        if self.slicer_type == "default":
            logger.info('')
            logger.info("Planar slicing using default function ...")
            self.layers = compas_slicer.slicers.create_planar_paths(self.mesh, planes)

        elif self.slicer_type == "meshcut":
            logger.info('')
            logger.info("Planar slicing using meshcut ...")
            self.layers = compas_slicer.slicers.create_planar_paths_meshcut(self.mesh, planes)

        elif self.slicer_type == "cgal":
            logger.info('')
            logger.info("Planar slicing using CGAL ...")
            self.layers = compas_slicer.slicers.create_planar_paths_cgal(self.mesh, planes)

        else:
            raise NameError("Invalid slicing type : " + self.slicer_type)


if __name__ == "__main__":
    pass



import numpy as np
from compas.geometry import Point
from compas_slicer.geometry import Path
from compas_slicer.geometry import Layer
import logging
import meshcut
import progressbar

logger = logging.getLogger('logger')

__all__ = ['create_planar_paths_meshcut']


def create_planar_paths_meshcut(mesh, planes):
    """Creates planar slices using the Meshcut library
    https://pypi.org/project/meshcut/ from Julien Rebetez

    Considers all resulting paths as CLOSED paths.
    Attention, this is a very slow method. We don't recommend using it.

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
        A compas mesh.
    planes: list, :class: 'compas.geometry.Plane'
    """

    # Convert compas mesh to meshcut mesh
    v = np.array(mesh.vertices_attributes('xyz'))
    vertices = v.reshape(-1, 3)  # vertices numpy array : #Vx3
    key_index = mesh.key_index()
    f = [[key_index[key] for key in mesh.face_vertices(fkey)] for fkey in mesh.faces()]
    faces = np.array(f)
    faces.reshape(-1, 3)  # faces numpy array : #Fx3
    vertices, faces = meshcut.merge_close_vertices(vertices, faces)
    meshcut_mesh = meshcut.TriangleMesh(vertices, faces)

    layers = []
    with progressbar.ProgressBar(max_value=len(planes)) as bar:
        for i, plane in enumerate(planes):
            paths_per_layer = []
            plane = meshcut.Plane(plane.point, plane.normal)  # define plane
            meshcut_array = meshcut.cross_section_mesh(meshcut_mesh, plane)

            for _j, item in enumerate(meshcut_array):
                # convert np array to list
                meshcut_list = item.tolist()

                points = [Point(p[0], p[1], p[2]) for p in meshcut_list]
                is_closed = True  # TODO is_closed is always set to True, has to be checked
                path = Path(points=points, is_closed=is_closed)
                paths_per_layer.append(path)

            layer = Layer(paths_per_layer)
            layers.append(layer)

            # advance progressbar
            bar.update(i)

    return layers


if __name__ == "__main__":
    pass



"""
********************************************************************************
geometry
********************************************************************************

.. currentmodule:: compas_slicer.geometry


Geometry in compas_slicer consists out of a Layer, Path, or Printpoint.
A Layer is generated when a geometry is sliced into layers and can therefore be
seen as a 'slice' of a geometry. Layers are typically organized horizontally,
but can also be organized vertically. A Layer consists out of one, or multiple
Paths (depending on the geometry).
A Path is a contour within a layer. A Path consists out of a list of
compas.Points, plus some additional attributes.
A PrintPoint consists out of a single compas.geometry.Point, with additional
functionality added for the printing process.


Classes
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    Layer
    Path
    PrintPoint
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .path import *  # noqa: F401 E402 F403
from .layer import *  # noqa: F401 E402 F403
from .print_point import *  # noqa: F401 E402 F403

__all__ = [name for name in dir() if not name.startswith('_')]



from compas_slicer.slicers.slice_utilities import ContoursBase
from compas.geometry import intersection_line_line_xy, distance_point_point_xy, scale_vector, add_vectors


class UVContours(ContoursBase):
    def __init__(self, mesh, p1, p2):
        ContoursBase.__init__(self, mesh)  # initialize from parent class
        self.p1 = p1  # tuple (u,v); first point in uv domain defining the cutting line
        self.p2 = p2  # tuple (u,v); second point in uv domain defining the cutting line

    def uv(self, vkey):
        return self.mesh.vertex[vkey]['uv']

    def edge_is_intersected(self, v1, v2):
        """ Returns True if the edge v1,v2 intersects the line in the uv domain, False otherwise. """
        p = intersection_line_line_xy((self.p1, self.p2), (self.uv(v1), self.uv(v2)))
        if p:
            if is_point_on_line_xy(p, (self.uv(v1), self.uv(v2))):
                if is_point_on_line_xy(p, (self.p1, self.p2)):
                    return True
        return False

    def find_zero_crossing_data(self, v1, v2):
        """ Finds the position of the zero-crossing on the edge u,v. """
        p = intersection_line_line_xy((self.p1, self.p2), (self.uv(v1), self.uv(v2)))
        d1, d2 = distance_point_point_xy(self.uv(v1), p), distance_point_point_xy(self.uv(v2), p)
        if d1 + d2 > 0:
            vec = self.mesh.edge_vector(v1, v2)
            vec = scale_vector(vec, d1 / (d1 + d2))
            pt = add_vectors(self.mesh.vertex_coordinates(v1), vec)
            return pt


# utility function

def is_point_on_line_xy(c, line, epsilon=1e-6):
    """
    Not using the equivalent function of compas, because for some reason it always returns True.

    c: list that represents a point with 2 coordinates [x.y] of [x,y,0]
    line: (p1, p2) where each pt represents a point with 2 coordinates [x.y] of [x,y,0]
    """
    a, b = line[0], line[1]
    cross_product = (c[1] - a[1]) * (b[0] - a[0]) - (c[0] - a[0]) * (b[1] - a[1])

    if abs(cross_product) > epsilon:
        return False

    dot_product = (c[0] - a[0]) * (b[0] - a[0]) + (c[1] - a[1]) * (b[1] - a[1])
    if dot_product < 0:
        return False

    squared_length_ba = (b[0] - a[0]) * (b[0] - a[0]) + (b[1] - a[1]) * (b[1] - a[1])
    if dot_product > squared_length_ba:
        return False

    return True


if __name__ == "__main__":
    pass



from compas.geometry import Frame, Point
from compas.geometry import Transformation
from compas.datastructures import mesh_bounding_box

import logging

logger = logging.getLogger('logger')

__all__ = ['move_mesh_to_point',
           'get_mid_pt_base']


def move_mesh_to_point(mesh, target_point):
    """Moves (translates) a mesh to a target point.

    Parameters
    ----------
    mesh: :class:`compas.datastructures.Mesh`
        A compas mesh.
    target_point: :class:`compas.geometry.Point`
        The point to move the mesh to.
    """
    mesh_center_pt = get_mid_pt_base(mesh)

    # transform mesh
    mesh_frame = Frame(mesh_center_pt, (1, 0, 0), (0, 1, 0))
    target_frame = Frame(target_point, (1, 0, 0), (0, 1, 0))

    T = Transformation.from_frame_to_frame(mesh_frame, target_frame)
    mesh.transform(T)

    logger.info("Mesh moved to: " + str(target_point))

    return mesh


def get_mid_pt_base(mesh):
    """Gets the middle point of the base (bottom) of the mesh.

    Parameters
    ----------
    mesh: :class:`compas.datastructures.Mesh`
        A compas mesh.

    Returns
    -------
    mesh_mid_pt: :class:`compas.geometry.Point`
        Middle point of the base of the mesh.

    """
    # get center bottom point of mesh model
    bbox = mesh_bounding_box(mesh)
    corner_pts = [bbox[0], bbox[2]]

    x = [p[0] for p in corner_pts]
    y = [p[1] for p in corner_pts]
    z = [p[2] for p in corner_pts]

    mesh_mid_pt = Point((sum(x) / 2), (sum(y) / 2), (sum(z) / 2))

    return mesh_mid_pt


if __name__ == "__main__":
    pass



__all__ = ['interpolation_slicing_default_param']


def interpolation_slicing_default_param(key):
    """ Returns the default parameters with the specified key. """
    if key in default_parameters:
        return default_parameters[key]
    else:
        raise ValueError('The parameter with key : ' + str(key) +
                         ' does not exist in the defaults of curved_slicing parameters. ')


default_parameters = \
    {
        'target_LOW_smooth_union': [False, 0],
        'target_LOW_geodesics_method': 'exact_igl',
        'target_HIGH_smooth_union': [False, 0],
        'target_HIGH_geodesics_method': 'exact_igl',
        'uneven_upper_targets_offset': 0,
    }



import compas
import numpy as np
import compas_slicer
from compas.datastructures import Mesh
from compas_slicer.utilities import utils
from compas.geometry import Polyline
from compas_slicer.geometry import Layer, VerticalLayer
from compas_slicer.post_processing import seams_align, unify_paths_orientation
import logging
import copy
from abc import abstractmethod
from compas.datastructures import mesh_bounding_box
from compas.geometry import distance_point_point_sqrd

logger = logging.getLogger('logger')

__all__ = ['BaseSlicer']


class BaseSlicer(object):
    """
    This is an organizational class that holds all the information for the slice process.
    Do not use this class directly in your python code. Instead use PlanarSlicer or InterpolationSlicer.
    This class is meant to be extended for the implementation of the various slicers.
    See :class:`compas.slicer.slicers.PlanarSlicer` and :class:`compas.slicer.slicers.InterpolationSlicer` as examples.

    Attributes
    ----------
    mesh: :class:`compas.datastructures.Mesh`
        Input mesh, has to be a triangular mesh (i.e. no quads or n-gons allowed)
    """

    def __init__(self, mesh):
        #  check input
        assert isinstance(mesh, compas.datastructures.Mesh), \
            "Input mesh must be of type <compas.datastructures.Mesh>, not " + str(type(mesh))
        utils.check_triangular_mesh(mesh)

        #  input
        self.mesh = mesh
        logger.info("Input Mesh with : %d vertices, %d Faces"
                    % (len(list(self.mesh.vertices())), len(list(self.mesh.faces()))))

        self.layer_height = None
        self.layers = []  # any class inheriting from Layer(horizontal sorting)

    ##############################
    #  --- Properties

    @property
    def number_of_points(self):
        """ Returns int: Total number of points in the slicer."""
        total_number_of_pts = 0
        for layer in self.layers:
            for path in layer.paths:
                total_number_of_pts += len(path.points)
        return total_number_of_pts

    @property
    def number_of_layers(self):
        """ Returns int: Total number of layers."""
        return len(self.layers)

    @property
    def number_of_paths(self):
        """ Returns tuple (int, int, int): Total number of paths, number of open paths, number of closed paths. """
        total_number_of_paths = 0
        closed_paths = 0
        open_paths = 0
        for layer in self.layers:
            total_number_of_paths += len(layer.paths)
            for path in layer.paths:
                if path.is_closed:
                    closed_paths += 1
                else:
                    open_paths += 1

        return total_number_of_paths, closed_paths, open_paths

    @property
    def vertical_layers(self):
        """ Returns a list of all the vertical layers stored in the slicer. """
        return [layer for layer in self.layers if isinstance(layer, VerticalLayer)]

    @property
    def horizontal_layers(self):
        """ Returns a list of all the layers stored in the slicer that are NOT vertical. """
        return [layer for layer in self.layers if not isinstance(layer, VerticalLayer)]

    ##############################
    #  --- Functions

    def slice_model(self, *args, **kwargs):
        """Slices the model and applies standard post-processing and removing of invalid paths."""

        self.generate_paths()
        self.remove_invalid_paths_and_layers()
        self.post_processing()

    @abstractmethod
    def generate_paths(self):
        """To be implemented by the inheriting classes. """
        pass

    def post_processing(self):
        """Applies standard post-processing operations: seams_align and unify_paths."""

        #  --- Align the seams between layers and unify orientation
        seams_align(self, align_with='x_axis')
        unify_paths_orientation(self)
        self.close_paths()

        logger.info("Created %d Layers with %d total number of points"
                    % (len(self.layers), self.number_of_points))

    def close_paths(self):
        """ For paths that are labeled as closed, it makes sure that the first and the last point are identical. """
        for layer in self.layers:
            for path in layer.paths:
                if path.is_closed:  # if the path is closed, first and last point should be the same.
                    if distance_point_point_sqrd(path.points[0], path.points[-1]) > 0.00001:  # if not already the same
                        path.points.append(path.points[0])

    def remove_invalid_paths_and_layers(self):
        """Removes invalid layers and paths from the slicer."""

        paths_to_remove = []
        layers_to_remove = []

        for i, layer in enumerate(self.layers):
            for j, path in enumerate(layer.paths):
                # check if a path has less than two points and appends to list to_remove
                if len(path.points) < 2:
                    paths_to_remove.append(path)
                    logger.warning("Invalid Path found: Layer %d, Path %d, %s" % (i, j, str(path)))
                    # check if the layer that the invalid path was in has only one path
                    # this means that path is now invalid, and the entire layer should be removed
                    if len(layer.paths) == 1:
                        layers_to_remove.append(layer)
                        logger.warning("Invalid Layer found: Layer %d, %s" % (i, str(layer)))
            # check for layers with less than one path and appends to list to_remove
            if len(layer.paths) < 1:
                layers_to_remove.append(layer)
                logger.warning("Invalid Layer found: Layer %d, %s" % (i, str(layer)))

        # compares the two lists and removes any invalid items
        for i, layer in enumerate(self.layers):
            for j, path in enumerate(layer.paths):
                if path in paths_to_remove:
                    layer.paths.remove(path)
            if layer in layers_to_remove:
                self.layers.remove(layer)

    def find_vertical_layers_with_first_path_on_base(self):
        bbox = mesh_bounding_box(self.mesh)
        z_min = min([p[2] for p in bbox])
        paths_on_base = []
        vertical_layer_indices = []
        d_threshold = 30

        for i, vertical_layer in enumerate(self.vertical_layers):
            first_path = vertical_layer.paths[0]
            avg_z_dist_from_min = np.average(np.array([abs(pt[2] - z_min) for pt in first_path.points]))

            if avg_z_dist_from_min < d_threshold:
                paths_on_base.append(vertical_layer.paths[0])
                vertical_layer_indices.append(i)

        return paths_on_base, vertical_layer_indices

    ##############################
    #  --- Output

    def printout_info(self):
        """Prints out information from the slicing process."""
        no_of_paths, closed_paths, open_paths = self.number_of_paths

        print("\n---- Slicer Info ----")
        print("Number of layers: %d" % self.number_of_layers)
        print("Number of paths: %d, open paths: %d, closed paths: %d" % (no_of_paths, open_paths, closed_paths))
        print("Number of sampling printpoints on layers: %d" % self.number_of_points)
        print("")

    def visualize_on_viewer(self, viewer, visualize_mesh=False, visualize_paths=True):
        """Visualizes slicing result using compas.viewers.

        Parameters
        ----------
        viewer: :class:`compas_viewers.objectviewer.Objectviewer`
            An instance of the Objectviewer class.
        visualize_mesh: bool, optional
            True to visualize mesh, False to not.
        visualize_paths: bool, optional
            True to visualize paths, False to not.
        """

        if visualize_mesh:
            viewer.add(self.mesh, settings={'color': '#ff0000',
                                            'opacity': 0.4, })

        if visualize_paths:
            for i, layer in enumerate(self.layers):
                for j, path in enumerate(layer.paths):
                    pts = copy.deepcopy(path.points)
                    if path.is_closed:
                        pts.append(pts[0])
                    polyline = Polyline(pts)
                    if isinstance(layer, compas_slicer.geometry.VerticalLayer):
                        viewer.add(polyline, name="VerticalSegment %d, Path %d" % (i, j),
                                   settings={'color': '#ffffff'})
                    else:
                        viewer.add(polyline, name="Layer %d, Path %d" % (i, j),
                                   settings={'color': '#ffffff'})

    ##############################
    #  --- To data, from data

    @classmethod
    def from_data(cls, data):
        """Construct a slicer from its data representation.

        Parameters
        ----------
        data: dict
            The data dictionary.

        Returns
        -------
        layer
            The constructed slicer.

        """
        mesh = Mesh.from_data(data['mesh'])
        slicer = cls(mesh)
        layers_data = data['layers']
        for layer_key in layers_data:
            if layers_data[layer_key]['layer_type'] == 'horizontal_layer':
                slicer.layers.append(Layer.from_data(layers_data[layer_key]))
            else:  # 'vertical_layer'
                slicer.layers.append(VerticalLayer.from_data(layers_data[layer_key]))
        slicer.layer_height = data['layer_height']
        return slicer

    def to_json(self, filepath, name):
        """Writes the slicer to a JSON file."""
        utils.save_to_json(self.to_data(), filepath, name)

    def to_data(self):
        """Returns a dictionary of structured data representing the data structure.

        Returns
        -------
        dict
            The slicer's data.

        """
        # To avoid errors when saving to Json, create a copy of the self.mesh and remove from it
        # any non-serializable attributes (by checking a random face and a random vertex, assuming
        # that all faces and vertices share the same types of attributes).
        mesh = self.mesh.copy()
        v_key = mesh.get_any_vertex()
        v_attrs = mesh.vertex_attributes(v_key)
        for attr_key in v_attrs:
            if not utils.is_jsonable(v_attrs[attr_key]):
                logger.error('vertex : ' + attr_key + str(v_attrs[attr_key]))
                for v in mesh.vertices():
                    mesh.unset_vertex_attribute(v, attr_key)

        f_key = mesh.get_any_face()
        f_attrs = mesh.face_attributes(f_key)
        for attr_key in f_attrs:
            if not utils.is_jsonable(f_attrs[attr_key]):
                logger.error('face : ' + attr_key, f_attrs[attr_key])
                mesh.update_default_face_attributes({attr_key: 0.0})  # just set all to 0.0

        # fill data dictionary with slicer info
        data = {'layers': self.get_layers_dict(),
                'mesh': mesh.to_data(),
                'layer_height': self.layer_height}
        return data

    def get_layers_dict(self):
        """Returns a dictionary consisting of the layers.
        """
        data = {}
        for i, layer in enumerate(self.layers):
            data[i] = layer.to_data()
        return data


if __name__ == "__main__":
    pass



import logging

from compas.geometry import Point
from compas.geometry import distance_point_point

logger = logging.getLogger('logger')

__all__ = ['seams_align']


def seams_align(slicer, align_with="next_path"):
    """Aligns the seams (start- and endpoint) of a print.

    Parameters
    ----------
    slicer: :class:`compas_slicer.slicers.BaseSlicer`
        An instance of one of the compas_slicer.slicers classes.
    align_with: str or :class:`compas.geometry.Point`
        Direction to orient the seams in.
        next_path    = orients the seam to the next path
        origin       = orients the seam to the origin (0,0,0)
        x_axis       = orients the seam to the x_axis
        y_axis       = orients the seam to the y_axis
        Point(x,y,z) = orients the seam according to the given point

    Returns
    -------
    None
    """
    #  TODO: Implement random seams
    logger.info("Aligning seams to: %s" % align_with)

    for i, layer in enumerate(slicer.layers):
        for j, path in enumerate(layer.paths):

            align_seams_for_current_path = path.is_closed  # should not happen if path is open

            if align_seams_for_current_path:
                #  get the points of the current layer and path
                path_to_change = layer.paths[j].points

                # check if start- and end-points are the same point
                if path_to_change[0] == path_to_change[-1]:
                    first_last_point_the_same = True
                    # if they are, remove the last point
                    path_to_change.pop(-1)
                else:
                    first_last_point_the_same = False

                if align_with == "next_path":
                    pt_to_align_with = None  # make sure aligning point is cleared

                    #  determines the correct point to align the current layer with
                    if len(layer.paths) == 1 and i == 0:
                        #  if ONE PATH and FIRST LAYER
                        #  >>> align with second layer
                        pt_to_align_with = slicer.layers[i + 1].paths[0].points[0]
                    if len(layer.paths) == 1 and i != 0:
                        #  if ONE PATH and NOT FIRST LAYER
                        #  >>> align with previous layer
                        pt_to_align_with = slicer.layers[i - 1].paths[0].points[0]
                    if len(layer.paths) != 1 and i == 0 and j == 0:
                        #  if MULTIPLE PATHS and FIRST LAYER and FIRST PATH
                        #  >>> align with second path of first layer
                        pt_to_align_with = slicer.layers[i].paths[i + 1].points[0]
                    if len(layer.paths) != 1 and j != 0:
                        #  if MULTIPLE PATHS and NOT FIRST PATH
                        #  >>> align with previous path
                        pt_to_align_with = slicer.layers[i].paths[j - 1].points[0]
                    if len(layer.paths) != 1 and i != 0 and j == 0:
                        #  if MULTIPLE PATHS and NOT FIRST LAYER and FIRST PATH
                        #  >>> align with first path of previous layer
                        pt_to_align_with = slicer.layers[i - 1].paths[j].points[0]

                elif align_with == "origin":
                    pt_to_align_with = Point(0, 0, 0)
                elif align_with == "x_axis":
                    pt_to_align_with = Point(2 ** 32, 0, 0)
                elif align_with == "y_axis":
                    pt_to_align_with = Point(0, 2 ** 32, 0)
                elif isinstance(align_with, Point):
                    pt_to_align_with = align_with
                else:
                    raise NameError("Unknown align_with : " + str(align_with))

                #  computes distance between pt_to_align_with and the current path points
                distance_current_pt_align_pt = [distance_point_point(pt_to_align_with, pt) for pt in path_to_change]
                #  gets the index of the closest point by looking for the minimum
                new_start_index = distance_current_pt_align_pt.index(min(distance_current_pt_align_pt))
                #  shifts the list by the distance determined
                shift_list = path_to_change[new_start_index:] + path_to_change[:new_start_index]
                #  shifts the list by the distance determined
                # layer.paths[j].points = shift_list

                if first_last_point_the_same:
                    shift_list = shift_list + [shift_list[0]]

                layer.paths[j].points = shift_list


if __name__ == "__main__":
    pass



"""
********************************************************************************
post_processing
********************************************************************************

Polyline simplification
=======================

.. autosummary::
    :toctree: generated/
    :nosignatures:

    simplify_paths_rdp

Sorting
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    seams_align
    seams_smooth
    sort_into_vertical_layers
    reorder_vertical_layers

Additional
==========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    generate_brim
    generate_raft
    spiralize_contours

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

#  Polyline simplification
from .simplify_paths_rdp import *  # noqa: F401 E402 F403

#  Sorting
from .seams_align import *  # noqa: F401 E402 F403
from .seams_smooth import *  # noqa: F401 E402 F403
from .sort_into_vertical_layers import *  # noqa: F401 E402 F403
from .reorder_vertical_layers import *  # noqa: F401 E402 F403

#  Orienting
from .unify_paths_orientation import *  # noqa: F401 E402 F403

#  Additional
from .generate_brim import *  # noqa: F401 E402 F403
from .generate_raft import *  # noqa: F401 E402 F403
from .spiralize_contours import *  # noqa: F401 E402 F403

__all__ = [name for name in dir() if not name.startswith('_')]



import os
import json
import rhinoscriptsyntax as rs
from compas.datastructures import Mesh
import Rhino.Geometry as rg
from compas_ghpython.artists import MeshArtist
from compas.geometry import Frame
from compas_ghpython.utilities import list_to_ghtree


#######################################
# --- Slicer

def load_slicer(path, folder_name, json_name):
    """ Loads slicer data. """
    data = load_json_file(path, folder_name, json_name)

    mesh = None
    paths_nested_list = []
    are_closed = []
    all_points = []

    if data:

        if 'mesh' in data:
            compas_mesh = Mesh.from_data(data['mesh'])
            artist = MeshArtist(compas_mesh)
            mesh = artist.draw()
        else:
            print('No mesh has been saved in the json file.')

        if 'layers' in data:
            layers_data = data['layers']

            for i in range(len(layers_data)):
                paths_nested_list.append([])  # save each layer on a different list
                layer_data = layers_data[str(i)]
                paths_data = layer_data['paths']

                for j in range(len(paths_data)):
                    path_data = paths_data[str(j)]
                    pts = []

                    are_closed.append(path_data['is_closed'])

                    for k in range(len(path_data['points'])):
                        pt = path_data['points'][str(k)]
                        pt = rs.AddPoint(pt[0], pt[1], pt[2])  # re-create points
                        pts.append(pt)
                    all_points.extend(pts)
                    path = rs.AddPolyline(pts)

                    # # create contour per layer
                    # try:
                    #     path = rs.AddPolyline(pts)
                    # except:
                    #     print('Attention! Could not add polyline at layer %d, path %d with %d points ' % (
                    #         i, j, len(path_data['points'])))

                    paths_nested_list[-1].append(path)
        else:
            print('No layers have been saved in the json file. Is this the correct json?')

    print('The slicer contains %d layers. ' % len(paths_nested_list))
    paths_nested_list = list_to_ghtree(paths_nested_list)
    return mesh, paths_nested_list, are_closed, all_points


#######################################
# --- Printpoints

def load_printpoints(path, folder_name, json_name):
    """ Loads a dict of compas_slicer printpoints. """
    data = load_json_file(path, folder_name, json_name)

    # geometry data
    points = []
    frames = []
    layer_heights = []
    up_vectors = []
    mesh_normals = []
    closest_support = []

    # fabrication related data
    velocities = []
    wait_times = []
    blend_radiuses = []
    extruder_toggles = []
    feasibility = []

    if data:
        for i in range(len(data)):
            data_point = data[str(i)]

            # geometry related data
            point = rg.Point3d(data_point["point"][0], data_point["point"][1], data_point["point"][2])
            points.append(point)

            compas_frame = Frame.from_data(data_point["frame"])
            pt, x_axis, y_axis = compas_frame.point, compas_frame.xaxis, compas_frame.yaxis
            frame = rs.PlaneFromNormal(pt, rs.VectorCrossProduct(rg.Vector3d(x_axis[0], x_axis[1], x_axis[2]),
                                                                 rg.Vector3d(y_axis[0], y_axis[1], y_axis[2])))
            frames.append(frame)

            layer_heights.append(data_point["layer_height"])

            v = data_point["up_vector"]
            up_vector = rg.Vector3d(v[0], v[1], v[2])
            up_vectors.append(up_vector)

            v = data_point["mesh_normal"]
            mesh_normal = rg.Vector3d(v[0], v[1], v[2])
            mesh_normals.append(mesh_normal)

            cp = data_point["closest_support_pt"]
            if cp:
                cp_pt = rg.Point3d(cp[0], cp[1], cp[2])
                closest_support.append(cp_pt)

            # fabrication related data
            velocities.append(data_point["velocity"])
            wait_times.append(data_point["wait_time"])
            blend_radiuses.append(data_point["blend_radius"])
            extruder_toggles.append(data_point["extruder_toggle"])
            feasibility.append(data_point["is_feasible"])

    return points, frames, layer_heights, up_vectors, mesh_normals, closest_support, velocities, wait_times, \
        blend_radiuses, extruder_toggles, feasibility


#######################################
# --- Lightweight path visualization

def lightweight_path_visualization(points, extruder_toggles, domain_start, domain_end, diameter, pipe_resolution):
    """ Visualize print paths with simple lines or pipes. """
    #  check input
    assert len(points) == len(extruder_toggles), \
        'Wrong length of input lists'

    print_path_pipes = []
    travel_path_lines = []

    domain_end = min(domain_end, len(points))  # make sure domain_end does not exceed len of pts

    points = points[domain_start:domain_end]
    extruder_toggles = extruder_toggles[domain_start:domain_end]
    for i in range(len(points) - 1):
        if extruder_toggles[i]:
            line = rg.Curve.CreateControlPointCurve([points[i], points[i + 1]])  # create line
            pipe = rg.Mesh.CreateFromCurvePipe(line, diameter / 2, pipe_resolution, 1, 0, True)
            print_path_pipes.append(pipe)
        else:
            line = rg.Curve.CreateControlPointCurve([points[i], points[i + 1]])
            travel_path_lines.append(line)  # add to travel path list

    return print_path_pipes, travel_path_lines


#######################################
# --- Render path visualization

def render_path_visualization(points, mesh_normals, layer_heights, up_vectors, extruder_toggles, cross_section,
                              planar_printing):
    """ Visualize print paths with simple loft surfaces. """

    # check input
    assert len(points) == len(mesh_normals) == len(layer_heights) == len(up_vectors) == len(extruder_toggles), \
        'Wrong length of input lists'

    loft_surfaces = []
    travel_path_lines = []

    if points[0] and mesh_normals[0] and layer_heights[0] and up_vectors[0]:  # check if any of the values are None

        if planar_printing:  # then make sure that all normals lie on the xy plane
            for n in mesh_normals:
                n[2] = 0

        # transform and scale cross sections accordingly
        cen = rs.CurveAreaCentroid(cross_section)[0]
        origin_plane = rg.Plane(cen, rg.Vector3d(1, 0, 0), rg.Vector3d(0, 0, 1))

        target_planes = []
        for i, pt in enumerate(points):
            target_plane = rg.Plane(pt, mesh_normals[i], up_vectors[i])
            target_planes.append(target_plane)

        cross_sections = []
        for h, target_plane in zip(layer_heights, target_planes):
            section = rs.ScaleObject(rs.CopyObject(cross_section), origin=cen, scale=[0.9 * h, 1, 0.9 * h])
            T = rg.Transform.PlaneToPlane(origin_plane, target_plane)
            rs.TransformObject(section, T)
            cross_sections.append(section)

        loft_surfaces = []
        travel_path_lines = []

        for i in range(len(points) - 1):
            if extruder_toggles[i]:
                loft = rs.AddLoftSrf([cross_sections[i], cross_sections[i + 1]])
                if loft:
                    loft_surfaces.append(loft[0])
            else:
                line = rg.Curve.CreateControlPointCurve([points[i], points[i + 1]])
                travel_path_lines.append(line)  # add to travel path list

    else:
        print('At least one of the inputs that you have provided are invalid. ')

    return loft_surfaces, travel_path_lines


#######################################
# --- Tool visualization

def tool_visualization(origin_coords, mesh, planes, i):
    """ Visualize example tool motion. """

    i = min(i, len(planes) - 1)  # make sure i doesn't go beyond available number of planes
    passed_path = None

    if planes[0]:
        origin = [float(origin_coords[0]), float(origin_coords[1]), float(origin_coords[2])]
        o = rg.Point3d(origin[0], origin[1], origin[2])
        x = rg.Vector3d(1, 0, 0)
        z = rg.Vector3d(0, 0, -1)

        ee_frame = rg.Plane(o, x, z)
        target_frame = planes[i]

        T = rg.Transform.PlaneToPlane(ee_frame, target_frame)
        mesh = rs.TransformObject(rs.CopyObject(mesh), T)

        passed_path = rs.AddPolyline([plane.Origin for plane in planes[:i + 1]])

    else:
        print('The planes you have provided are invalid. ')

    return mesh, passed_path


#######################################
# --- Create_targets (Curved slicing)

def create_targets(mesh, targets, resolution_mult, path, folder_name, json_name):
    """ Creation of targets for curved slicing. """

    avg_face_area = max(rs.MeshArea([mesh])) / rs.MeshFaceCount(mesh)
    div_num = max(20, int(resolution_mult * avg_face_area))

    pts = []
    for target in targets:
        print(div_num)
        pts.extend(rs.DivideCurve(target, div_num))

    vs = rs.MeshVertices(mesh)
    vertices = []
    vertex_indices = []
    for p in pts:
        closest_vi = get_closest_point_index(p, vs)
        if closest_vi not in vertex_indices:
            ds_from_targets = [distance_of_pt_from_crv(vs[closest_vi], target) for target in targets]
            if min(ds_from_targets) < 0.1:  # hardcoded threshold value
                vertices.append(vs[closest_vi])
                vertex_indices.append(closest_vi)

    save_json_file(vertex_indices, path, folder_name, json_name)
    return pts, vertices, vertex_indices


#######################################
# --- Create_targets (Curved slicing)

def load_multiple_meshes(starts_with, ends_with, path, folder_name):
    """ Load all the meshes that have the specified name, and print them in different colors. """
    filenames = get_files_with_name(starts_with, ends_with, os.path.join(path, folder_name, 'output'))
    meshes = [Mesh.from_obj(os.path.join(path, folder_name, 'output', filename)) for filename in filenames]

    loaded_meshes = []
    for i, m in enumerate(meshes):
        artist = MeshArtist(m)
        color = get_color(i, total=len(meshes))
        mesh = artist.draw(color)
        loaded_meshes.append(mesh)

    return loaded_meshes


#######################################
# --- Load json points

def load_json_points(path, folder_name, json_name):
    """
    Loads a json file that stores a dictionary of N points in the format:
    data['0']=[x0,y0,z0], ...,  data['N']=[xN,yN,zN]
    """
    data = load_json_file(path, folder_name, json_name)
    points = None

    if data:
        points = []
        for i in range(len(data)):
            points.append(rg.Point3d(data[str(i)][0], data[str(i)][1], data[str(i)][2]))
    return points


#######################################
# --- Load json vectors

def load_json_vectors(path, folder_name, json_name):
    """
    Loads a json file from the 'output' folder,
    that stores a vector field in the format:
    data['0']=[x0,y0,z0], ...,  data['N']=[xN,yN,zN]
    """
    data = load_json_file(path, folder_name, json_name)
    vectors = None

    if data:
        vectors = []
        for i in range(len(data)):
            vectors.append(rg.Vector3d(data[str(i)][0], data[str(i)][1], data[str(i)][2]))
    return vectors


#######################################
# --- Load json polylines

def load_json_polylines(path, folder_name, json_name):
    """
    Loads a json file that stores a dictionary of N polylines in the format:
    data['0']=points_dict_0, ..., data['N'] = points_dict_N, where points_dict is in the format:
    points_dict['0']=[x0,y0,z0], ...,  points_dict['N']=[xN,yN,zN]
    """
    data = load_json_file(path, folder_name, json_name)
    all_points = []
    polylines = []

    if data:
        for i in range(len(data)):
            pts = []
            pts_dict = data[str(i)]

            for j in range(len(pts_dict)):
                p = pts_dict[str(j)]
                pts.append(rg.Point3d(p[0], p[1], p[2]))
            all_points.extend(pts)
            polylines.append(rs.AddPolyline(pts))

    return polylines, all_points


#######################################
# --- Load json BaseBoundaries

def load_base_boundaries(path, folder_name, json_name):
    """
    Loads a json file that stores a dictionary of BaseBoundary classes in the format:
    data['0']=base_boundary_1.to_data(), ..., data['N']=base_boundary_N.to_data()
    """
    data = load_json_file(path, folder_name, json_name)
    points, vectors, number_of_boundaries = [], [], None

    if data:
        number_of_boundaries = len(data)

        for i in range(len(data)):
            p = data[str(i)]['points']
            v = data[str(i)]['up_vectors']

            points.extend([rg.Point3d(p[key][0], p[key][1], p[key][2]) for key in p])
            vectors.extend([rg.Vector3d(v[key][0], v[key][1], v[key][2]) for key in v])

    return points, vectors, number_of_boundaries


#######################################
# --- Load json BaseBoundaries

def distance_fields_interpolation(path, folder_name, json_name1, json_name2, weight):
    """ Simple interpolation of the distance fields that are stored in the two json files. """
    distances_LOW = load_json_file(path, folder_name, json_name1)
    distances_HIGH = load_json_file(path, folder_name, json_name2)

    if distances_LOW and distances_HIGH:
        assert (len(distances_LOW) == len(distances_HIGH)), 'Wrong number of distances provided. '
        return [d2 * weight - d1 * (1 - weight) for d1, d2 in zip(distances_LOW, distances_HIGH)]


#######################################
# --- Load json BaseBoundaries

def distance_fields_weighted_interpolation(path, folder_name, json_name, weight):
    """ Weighted interpolation of the two distance fields in the json file. """
    data = load_json_file(path, folder_name, json_name)
    interpolation = []

    if data:
        distances_LOW = data['distances_LOW']
        distances_HIGH_dict = data['distances_HIGH_clusters']
        t_ends_HIGH = data['t_end_per_cluster_HIGH']
        has_blend_union = data['has_blend_union_HIGH']
        blend_radius = data['blend_radius']

        weights_remapped = [remap_unbound(weight, 0, t_end, 0, 1) for t_end in t_ends_HIGH]

        for i in range(len(distances_LOW)):
            d_low = distances_LOW[i]
            ds_high = [distances_HIGH_dict[str(j)][i] for j in range(len(distances_HIGH_dict))]

            ds = [(w - 1) * d_low + w * d_high for d_high, w in zip(ds_high, weights_remapped)]

            if has_blend_union:
                d_final = blend_union_list(ds, blend_radius)
            else:
                d_final = min(ds)
            interpolation.append(abs(d_final))

    return interpolation


##############################################
# --- utilities
##############################################

def missing_input():
    """ Deals with cases where the user has not defined all the necessary inputs. """
    print('Please provide all the inputs')


def load_json_file(path, folder_name, json_name, in_output_folder=True):
    """ Loads data from json. """

    if in_output_folder:
        filename = os.path.join(os.path.join(path), folder_name, 'output', json_name)
    else:
        filename = os.path.join(os.path.join(path), folder_name, json_name)
    data = None

    if os.path.isfile(filename):
        with open(filename, 'r') as f:
            data = json.load(f)
        print("Loaded Json: '" + filename + "'")
    else:
        print("Attention! Filename: '" + filename + "' does not exist. ")

    return data


def save_json_file(data, path, folder_name, json_name):
    """ Saves data to json. """
    filename = os.path.join(path, folder_name, json_name)
    with open(filename, 'w') as f:
        f.write(json.dumps(data, indent=3, sort_keys=True))
    print("Saved to Json: '" + filename + "'")


def get_closest_point_index(pt, pts):
    """ Closest point index of the pts from pt. """
    distances = [rs.Distance(p, pt) for p in pts]
    min_index = distances.index(min(distances))
    return min_index


def distance_of_pt_from_crv(pt, crv):
    """ Smallest distance from point to curve. """
    param = rs.CurveClosestPoint(crv, pt)
    cp = rs.EvaluateCurve(crv, param)
    return rs.Distance(pt, cp)


def get_files_with_name(startswith, endswith, DATA_PATH):
    """ Find all files with the specified start and end in the data path. """
    files = []
    for file in os.listdir(DATA_PATH):
        if file.startswith(startswith) and file.endswith(endswith):
            files.append(file)
    print('Found %d files with the given criteria : ' % len(files) + str(files))
    return files


def get_color(i, total):
    """ Returns a color per index interpolating the colorspace of 5 colors that are hardcoded (c1 .. c5). """
    i, total = float(i), float(total)

    c1 = rg.Vector3d(234, 38, 0.0)  # 0.00
    c2 = rg.Vector3d(234, 126, 0.0)  # 0.25
    c3 = rg.Vector3d(254, 244, 84)  # 0.50
    c4 = rg.Vector3d(173, 203, 249)  # 0.75
    c5 = rg.Vector3d(75, 107, 169)  # 1.00

    a = (i / (total - 1)) * 4.0
    if 0.0 <= a < 1:
        c_left, c_right = c1, c2
    elif 1 <= a < 2:
        c_left, c_right = c2, c3
        a -= 1
    elif 2 <= a < 3:
        c_left, c_right = c3, c4
        a -= 2
    else:
        c_left, c_right = c4, c5
        a -= 3

    weight = a
    c = (1 - weight) * c_left + weight * c_right
    return [int(c[0]), int(c[1]), int(c[2])]


def remap_unbound(input_val, in_from, in_to, out_from, out_to):
    """ Remap numbers without clamping values. """
    out_range = out_to - out_from
    in_range = in_to - in_from
    in_val = input_val - in_from
    val = (float(in_val) / in_range) * out_range
    out_val = out_from + val
    return out_val


def blend_union_list(values, r):
    """ Returns a blend union of the elements of the list, with blend radius r. """
    d_result = 9999999  # very big number
    for d in values:
        d_result = blend_union(d_result, d, r)
    return d_result


def blend_union(da, db, r):
    """ Blend union of the distances da, db with blend radius r. """
    e = max(r - abs(da - db), 0)
    return min(da, db) - e * e * 0.25 / r


if __name__ == "__main__":
    pass



import logging
from compas_slicer.utilities import find_next_printpoint

logger = logging.getLogger('logger')

__all__ = ['set_wait_time',
           'override_wait_time']


def set_wait_time(print_organizer, wait_type, wait_time):
    """Automatically sets a wait time for the printpoints.

    Sets a wait time for the printpoints, either before extrusion starts,
    after extrusion finishes, or in both cases.

    Parameters
    ----------
    print_organizer: :class:`compas_slicer.print_organization.BasePrintOrganizer`
    wait_type: str
        wait_before_extrusion:  sets a wait time before extrusion (extruder_toggle False to True)
        wait_after_extrusion: sets a wait time after extrusion (extruder_toggle True to False)
        wait_before_and_after_extrusion: sets a wait time before, and after extrusion
    wait_time: float
        Time in seconds to introduce to add as a wait time
    """

    pp_dict = print_organizer.printpoints_dict

    for layer_key in pp_dict:
        for path_key in pp_dict[layer_key]:
            for pp in pp_dict[layer_key][path_key]:
                assert pp.extruder_toggle is not None, \
                    'You need to set the extruder toggles first, before you can automatically set the wait time'

    logger.info("Setting wait time")

    for i, layer_key in enumerate(pp_dict):
        for j, path_key in enumerate(pp_dict[layer_key]):
            for k, printpoint in enumerate(pp_dict[layer_key][path_key]):
                # compares the current point with the next point
                next_ppt = find_next_printpoint(pp_dict, layer_key, path_key, i, j, k)

                # for the brim layer don't add any wait times
                if not print_organizer.slicer.layers[i].is_brim and next_ppt:
                    if wait_type == "wait_before_extrusion":
                        if printpoint.extruder_toggle is False and next_ppt.extruder_toggle is True:
                            next_ppt.wait_time = wait_time
                    elif wait_type == "wait_after_extrusion":
                        if printpoint.extruder_toggle is True and next_ppt.extruder_toggle is False:
                            next_ppt.wait_time = wait_time
                    elif wait_type == "wait_before_and_after_extrusion":
                        if printpoint.extruder_toggle is False and next_ppt.extruder_toggle is True:
                            next_ppt.wait_time = wait_time
                        if printpoint.extruder_toggle is True and next_ppt.extruder_toggle is False:
                            next_ppt.wait_time = wait_time


def override_wait_time(print_organizer, override_value):
    """Overrides the wait_time value for the printpoints with a user-defined value.

    Parameters
    ----------
    print_organizer: :class:`compas_slicer.print_organization.BasePrintOrganizer`
    override_value: float
        Value to override the wait_time values with.
    """

    pp_dict = print_organizer.printpoints_dict

    for layer_key in pp_dict:
        for path_key in pp_dict[layer_key]:
            path_printpoints = pp_dict[layer_key][path_key]
            for printpoint in path_printpoints:
                printpoint.wait_time = override_value


if __name__ == "__main__":
    pass



import logging
import math

import compas_slicer
from compas_slicer.geometry import Layer
from compas_slicer.geometry import Path

from compas.geometry import Point
from compas.geometry import Line
from compas.geometry import Vector
from compas.geometry import bounding_box_xy
from compas.geometry import offset_polygon
from compas.geometry import intersection_line_line
from compas.geometry import offset_line

logger = logging.getLogger('logger')

__all__ = ['generate_raft']


def generate_raft(slicer,
                  raft_offset=10,
                  distance_between_paths=10,
                  direction="horizontal",
                  raft_layers=1,
                  raft_layer_height="default"):
    """Creates a raft.

    Parameters
    ----------
    slicer: :class:`compas_slicer.slicers.BaseSlicer`
        An instance of one of the compas_slicer.slicers.BaseSlicer classes.
    raft_offset: float
        Distance (in mm) that the raft should be offsetted from the first layer. Defaults to 10mm
    distance_between_paths: float
        Distance (in mm) between the printed lines of the raft. Defaults to 10mm
    direction: str
        horizontal: Create a raft in the horizontal direction
        vertical: Create a raft in the vertical direction
        diagonal: Create a raft int the diagonal direction
    raft_layers: int
        Number of raft layers to add. Defaults to 1
    raft_layer_height: float
        Layer height of the raft layers. Defaults to same value as used in the slicer.
    """

    # check if a raft_layer_height is specified, if not, use the slicer.layer_height value
    if raft_layer_height == "default":
        raft_layer_height = slicer.layer_height

    logger.info("Generating raft")

    # find if slicer has vertical or horizontal layers, and select which paths are to be offset.
    if isinstance(slicer.layers[0], compas_slicer.geometry.VerticalLayer):  # Vertical layers
        # then find all paths that lie on the print platform and make them brim.
        paths_to_offset, _ = slicer.find_vertical_layers_with_first_path_on_base()

    else:  # Horizontal layers
        # then replace the first layer with a raft layer.
        paths_to_offset = slicer.layers[0].paths

    # get flat lists of points in bottom layer
    all_pts = []
    for path in paths_to_offset:
        for pt in path.points:
            all_pts.append(pt)

    # get xy bounding box of bottom layer and create offset
    bb_xy = bounding_box_xy(all_pts)
    bb_xy_offset = offset_polygon(bb_xy, -raft_offset)
    # bring points in the xy_offset to the correct height
    for pt in bb_xy_offset:
        pt[2] = slicer.layers[0].paths[0].points[0][2]

    # calculate x range, y range, and number of steps
    x_range = abs(bb_xy_offset[0][0] - bb_xy_offset[1][0])
    y_range = abs(bb_xy_offset[0][1] - bb_xy_offset[3][1])

    # get maximum values of the bounding box
    bb_max_x_right = bb_xy_offset[1][0]
    bb_max_y_top = bb_xy_offset[3][1]

    # get point in bottom left corner as raft start point
    raft_start_pt = Point(bb_xy_offset[0][0], bb_xy_offset[0][1], bb_xy_offset[0][2])

    # create starting line for diagonal direction
    if direction == "diagonal":
        c = math.sqrt(2*(distance_between_paths**2))

        pt1 = Point(raft_start_pt[0] + c, raft_start_pt[1], raft_start_pt[2])
        pt2 = Point(pt1[0] - y_range, pt1[1] + y_range, pt1[2])
        line = Line(pt1, pt2)

    # move all points in the slicer up so that raft layers can be inserted
    for i, layer in enumerate(slicer.layers):
        for j, path in enumerate(layer.paths):
            for k, pt in enumerate(path.points):
                slicer.layers[i].paths[j].points[k] = Point(pt[0], pt[1], pt[2] + (raft_layers)*raft_layer_height)

    for i in range(raft_layers):

        iter = 0
        raft_points = []

        # create raft points depending on the chosen direction
        while True and iter < 9999:  # to avoid infinite while loop in case something is not correct
            # ===============
            # VERTICAL RAFT
            # ===============
            if direction == "vertical":
                raft_pt1 = Point(raft_start_pt[0] + iter*distance_between_paths, raft_start_pt[1], raft_start_pt[2] + i*raft_layer_height)
                raft_pt2 = Point(raft_start_pt[0] + iter*distance_between_paths, raft_start_pt[1] + y_range, raft_start_pt[2] + i*raft_layer_height)

                if raft_pt2[0] > bb_max_x_right or raft_pt1[0] > bb_max_x_right:
                    break

            # ===============
            # HORIZONTAL RAFT
            # ===============
            elif direction == "horizontal":
                raft_pt1 = Point(raft_start_pt[0], raft_start_pt[1] + iter*distance_between_paths, raft_start_pt[2] + i*raft_layer_height)
                raft_pt2 = Point(raft_start_pt[0] + x_range, raft_start_pt[1] + iter*distance_between_paths, raft_start_pt[2] + i*raft_layer_height)

                if raft_pt2[1] > bb_max_y_top or raft_pt1[1] > bb_max_y_top:
                    break

            # ===============
            # DIAGONAL RAFT
            # ===============
            elif direction == "diagonal":
                # create offset of the initial diagonal line
                offset_l = offset_line(line, iter*distance_between_paths, Vector(0, 0, -1))

                # get intersections for the initial diagonal line with the left and bottom of the bb
                int_left = intersection_line_line(offset_l, [bb_xy_offset[0], bb_xy_offset[3]])
                int_bottom = intersection_line_line(offset_l, [bb_xy_offset[0], bb_xy_offset[1]])

                # get the points at the intersections
                raft_pt1 = Point(int_left[0][0], int_left[0][1], int_left[0][2] + i*raft_layer_height)
                raft_pt2 = Point(int_bottom[0][0], int_bottom[0][1], int_bottom[0][2] + i*raft_layer_height)

                # if the intersection goes beyond the height of the left side of the bounding box:
                if int_left[0][1] > bb_max_y_top:
                    # create intersection with the top side
                    int_top = intersection_line_line(offset_l, [bb_xy_offset[3], bb_xy_offset[2]])
                    raft_pt1 = Point(int_top[0][0], int_top[0][1], int_top[0][2] + i*raft_layer_height)

                    # if intersection goes beyond the length of the top side, break
                    if raft_pt1[0] > bb_max_x_right:
                        break

                # if the intersection goes beyond the length of the bottom side of the bounding box:
                if int_bottom[0][0] > bb_max_x_right:
                    # create intersection with the right side
                    int_right = intersection_line_line(offset_l, [bb_xy_offset[1], bb_xy_offset[2]])
                    raft_pt2 = Point(int_right[0][0], int_right[0][1], int_right[0][2] + i*raft_layer_height)

                    # if intersection goes beyond the height of the right side, break
                    if raft_pt2[1] > bb_xy_offset[2][1]:
                        break

            # append to list alternating
            if iter % 2 == 0:
                raft_points.extend((raft_pt1, raft_pt2))
            else:
                raft_points.extend((raft_pt2, raft_pt1))

            iter += 1

        # create raft layer
        raft_layer = Layer([Path(raft_points, is_closed=False)])
        raft_layer.is_raft = True
        # insert raft layer in the correct position into the slicer
        slicer.layers.insert(i, raft_layer)


if __name__ == "__main__":
    pass



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .mesh_attributes_handling import *  # noqa: F401 E402 F403
from .compound_target import *  # noqa: F401 E402 F403
from .geodesics import *  # noqa: F401 E402 F403
from .assign_vertex_distance import *  # noqa: F401 E402 F403
from .gradient import *  # noqa: F401 E402 F403
from .region_split import *  # noqa: F401 E402 F403

__all__ = [name for name in dir() if not name.startswith('_')]



import logging
from compas.geometry import Point

logger = logging.getLogger('logger')

__all__ = ['spiralize_contours']


def spiralize_contours(slicer):
    """Spiralizes contours. Only supports a constant layer height.
    Can only be used for geometries consisting out of a single contour (vases).

    Parameters
    ----------
    slicer: :class: 'compas_slicer.slicers.PlanarSlicer'
        An instance of the compas_slicer.slicers.PlanarSlicer class.
    """
    # retrieves layer height by subtracting z of first point of layer 1 from layer 0
    layer_height = slicer.layers[1].paths[0].points[0][2] - slicer.layers[0].paths[0].points[0][2]

    for i, layer in enumerate(slicer.layers):
        if len(layer.paths) == 1:
            for path in layer.paths:
                for i, point in enumerate(path.points):
                    # get the number of points in a layer
                    no_of_points = len(path.points)
                    # calculates distance to move
                    distance_to_move = layer_height / no_of_points
                    # adds the distance to move to the z value and create new points
                    path.points[i] = Point(point[0], point[1], point[2] + (i*distance_to_move))
                # removes the first item to create a smooth transition to the next layer
                path.points.pop(0)
        else:
            logger.warning("Spiralize contours only works for layers consisting out of a single path, contours were not changed, spiralize contour skipped for layer %i" % i)


if __name__ == "__main__":
    pass



from compas_slicer.geometry import Path
from compas_slicer.geometry import Layer
import logging
import progressbar
from compas.geometry import intersection_segment_plane
from compas_slicer.slicers.slice_utilities import ContoursBase

logger = logging.getLogger('logger')

__all__ = ['create_planar_paths']


def create_planar_paths(mesh, planes):
    """
    Creates planar contours. Does not rely on external libraries.
    It is currently the only method that can return identify OPEN versus CLOSED paths.

    Parameters
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
        The mesh to be sliced
    planes: list, :class: 'compas.geometry.Plane'
    """

    layers = []

    with progressbar.ProgressBar(max_value=len(planes)) as bar:
        for i, plane in enumerate(planes):

            intersection = PlanarContours(mesh, plane)
            intersection.compute()

            paths = []
            if len(intersection.sorted_point_clusters) > 0:
                for key in intersection.sorted_point_clusters:
                    is_closed = intersection.closed_paths_booleans[key]
                    path = Path(points=intersection.sorted_point_clusters[key], is_closed=is_closed)
                    paths.append(path)

                layers.append(Layer(paths))

            bar.update(i)

    return layers


class PlanarContours(ContoursBase):
    """
    Finds the iso-contours of the function f(x) = vertex_coords.z - plane.z
    on the mesh.

    Attributes
    ----------
    mesh: :class: 'compas.datastructures.Mesh'
    plane: list, :class: 'compas.geometry.Plane'
    """
    def __init__(self, mesh, plane):
        self.plane = plane
        ContoursBase.__init__(self, mesh)  # initialize from parent class

    def edge_is_intersected(self, u, v):
        """ Returns True if the edge u,v has a zero-crossing, False otherwise. """
        a = self.mesh.vertex_attributes(u, 'xyz')
        b = self.mesh.vertex_attributes(v, 'xyz')
        z = [a[2], b[2]]  # check if the plane.z is withing the range of [a.z, b.z]
        return min(z) <= self.plane.point[2] < max(z)

    def find_zero_crossing_data(self, u, v):
        """ Finds the position of the zero-crossing on the edge u,v. """
        a = self.mesh.vertex_attributes(u, 'xyz')
        b = self.mesh.vertex_attributes(v, 'xyz')
        return intersection_segment_plane((a, b), self.plane)


if __name__ == "__main__":
    pass



import logging
import compas_slicer
import compas_slicer.utilities.utils as utils
import numpy as np
from compas_slicer.geometry import Path

logger = logging.getLogger('logger')

__all__ = ['Layer',
           'VerticalLayer',
           'VerticalLayersManager']


class Layer(object):
    """
    A Layer stores a group of ordered paths that are generated when a geometry is sliced.
    Layers are typically organized horizontally, but can also be organized vertically (see VerticalLayer).
    A Layer consists of one, or multiple Paths (depending on the geometry).

    Attributes
    ----------
    paths: list
        :class:`compas_slicer.geometry.Path`
    is_brim: bool
        True if this layer is a brim layer.
    number_of_brim_offsets: int
        The number of brim offsets this layer has (None if no brim).
    is_raft: bool
        True if this layer is a raft layer.
    """

    def __init__(self, paths):
        # check input
        if paths is None:
            paths = []
        if len(paths) > 0:
            assert isinstance(paths[0], compas_slicer.geometry.Path)
        self.paths = paths

        self.min_max_z_height = (None, None)  # Tuple containing the min and max z height of the layer.
        if paths:
            self.calculate_z_bounds()

        # brim
        self.is_brim = False
        self.number_of_brim_offsets = None

        # raft
        self.is_raft = False

    def __repr__(self):
        no_of_paths = len(self.paths) if self.paths else 0
        return "<Layer object with %i paths>" % no_of_paths

    @property
    def total_number_of_points(self):
        """Returns the total number of points within the layer."""
        num = 0
        for path in self.paths:
            num += len(path.printpoints)
        return num

    def calculate_z_bounds(self):
        """ Fills in the attribute self.min_max_z_height. """
        assert len(self.paths) > 0, "You cannot calculate z_bounds because the list of paths is empty."
        z_min = 2 ** 32  # very big number
        z_max = -2 ** 32  # very small number
        for path in self.paths:
            for pt in path.points:
                z_min = min(z_min, pt[2])
                z_max = max(z_max, pt[2])
        self.min_max_z_height = (z_min, z_max)

    @classmethod
    def from_data(cls, data):
        """Construct a layer from its data representation.

        Parameters
        ----------
        data: dict
            The data dictionary.

        Returns
        -------
        layer
            The constructed layer.
        """
        paths_data = data['paths']
        paths = [Path.from_data(paths_data[key]) for key in paths_data]
        layer = cls(paths=paths)
        layer.is_brim = data['is_brim']
        layer.number_of_brim_offsets = data['number_of_brim_offsets']
        layer.min_max_z_height = data['min_max_z_height']
        return layer

    def to_data(self):
        """Returns a dictionary of structured data representing the data structure.

        Returns
        -------
        dict
            The layer's data.
        """
        data = {'paths': {i: [] for i in range(len(self.paths))},
                'layer_type': 'horizontal_layer',
                'is_brim': self.is_brim,
                'number_of_brim_offsets': self.number_of_brim_offsets,
                'min_max_z_height': self.min_max_z_height}
        for i, path in enumerate(self.paths):
            data['paths'][i] = path.to_data()
        return data


class VerticalLayer(Layer):
    """
    Vertical ordering. A VerticalLayer stores the print paths sorted in vertical groups.
    It is created with an empty list of paths that is filled in afterwards.

    Attributes
    ----------
    id: int
        Identifier of vertical layer.
    """

    def __init__(self, id=0, paths=None):
        Layer.__init__(self, paths=paths)
        self.id = id
        self.head_centroid = None

    def __repr__(self):
        no_of_paths = len(self.paths) if self.paths else 0
        return "<Vertical Layer object with id : %d and %d paths>" % (self.id, no_of_paths)

    def append_(self, path):
        """ Add path to self.paths list. """
        self.paths.append(path)
        self.compute_head_centroid()
        self.calculate_z_bounds()

    def compute_head_centroid(self):
        """ Find the centroid of all the points of the last path in the self.paths list"""
        pts = np.array(self.paths[-1].points)
        self.head_centroid = np.mean(pts, axis=0)

    def printout_details(self):
        """ Prints the details of the class. """
        logger.info("VerticalLayer id : %d" % self.id)
        logger.info("Total number of paths : %d" % len(self.paths))

    def to_data(self):
        """Returns a dictionary of structured data representing the data structure.

        Returns
        -------
        dict
            The vertical layer's data.
        """
        data = {'paths': {i: [] for i in range(len(self.paths))},
                'min_max_z_height': self.min_max_z_height,
                'layer_type': 'vertical_layer'}
        for i, path in enumerate(self.paths):
            data['paths'][i] = path.to_data()
        return data

    @classmethod
    def from_data(cls, data):
        """Construct a vertical layer from its data representation.

        Parameters
        ----------
        data: dict
            The data dictionary.

        Returns
        -------
        layer
            The constructed vertical layer.
        """
        paths_data = data['paths']
        paths = [Path.from_data(paths_data[key]) for key in paths_data]
        layer = cls(id=None)
        layer.paths = paths
        layer.min_max_z_height = data['min_max_z_height']
        return layer


class VerticalLayersManager:
    """
    Creates empty vertical layers and assigns to the input paths to the fitting vertical layer using the add() function.
    The criterion for grouping paths to VerticalLayers is based on the proximity of the centroids of the paths.
    If the input paths don't fit in any vertical layer, then new vertical layer is created with that path.

    Attributes
    ----------
    threshold_max_centroid_dist:
        float. The maximum get_distance that the centroids of two successive paths can have to belong in the same
        VerticalLayer.
    max_paths_per_layer: int
        Maximum number of layers that a vertical layer can consist of.
        If None, then the vertical layer has an unlimited number of layers.
    """

    def __init__(self, threshold_max_centroid_dist=25.0, max_paths_per_layer=None):
        self.layers = [VerticalLayer(id=0)]  # vertical_layers_print_data that contain isocurves (compas_slicer.Path)
        self.threshold_max_centroid_dist = threshold_max_centroid_dist
        self.max_paths_per_layer = max_paths_per_layer

    def add(self, path):
        selected_layer = None

        #  Find an eligible layer for path (called selected_layer)
        if len(self.layers[0].paths) == 0:  # first path goes to first layer
            selected_layer = self.layers[0]

        else:  # find the candidate segment for new isocurve
            centroid = np.mean(np.array(path.points), axis=0)
            other_centroids = get_vertical_layers_centroids_list(self.layers)
            candidate_layer = self.layers[utils.get_closest_pt_index(centroid, other_centroids)]

            if np.linalg.norm(candidate_layer.head_centroid - centroid) < self.threshold_max_centroid_dist:
                if self.max_paths_per_layer:
                    if len(candidate_layer.paths) < self.max_paths_per_layer:
                        selected_layer = candidate_layer
                else:
                    selected_layer = candidate_layer

            if not selected_layer:  # then create new layer
                selected_layer = VerticalLayer(id=self.layers[-1].id + 1)
                self.layers.append(selected_layer)

        selected_layer.append_(path)


def get_vertical_layers_centroids_list(vert_layers):
    """ Returns a list with points that are the centroids of the heads of all vertical_layers_print_data. The head
    of a vertical_layer is its last path. """
    head_centroids = []
    for vert_layer in vert_layers:
        head_centroids.append(vert_layer.head_centroid)
    return head_centroids


if __name__ == "__main__":
    pass



from compas.geometry import norm_vector, Vector
import logging

logger = logging.getLogger('logger')

__all__ = ['set_blend_radius']


def set_blend_radius(print_organizer, d_fillet=10, buffer=0.3):
    """Sets the blend radius (filleting) for the robotic motion.

    Parameters
    ----------
    print_organizer: :class:`compas_slicer.slicers.BasePrintOrganizer`
    d_fillet: float
        Value to attempt to fillet with. Defaults to 10 mm.
    buffer: float
        Buffer to make sure that the blend radius is never too big.
        Defaults to 0.3.
    """

    logger.info("Setting blend radius")
    pp_dict = print_organizer.printpoints_dict

    for layer_key in pp_dict:
        for path_key in pp_dict[layer_key]:
            path_printpoints = pp_dict[layer_key][path_key]
            for i, printpoint in enumerate(path_printpoints):
                neighboring_items = print_organizer.get_printpoint_neighboring_items(layer_key, path_key, i)

                radius = d_fillet
                if neighboring_items[0]:
                    radius = min(radius,
                                 norm_vector(Vector.from_start_end(neighboring_items[0].pt, printpoint.pt)) * buffer)

                if neighboring_items[1]:
                    radius = min(radius,
                                 norm_vector(Vector.from_start_end(neighboring_items[1].pt, printpoint.pt)) * buffer)

                radius = round(radius, 5)

                printpoint.blend_radius = radius


if __name__ == "__main__":
    pass



import logging
from compas_slicer.parameters import get_param

logger = logging.getLogger('logger')

__all__ = ['create_gcode_text']


def create_gcode_text(printpoints_dict, parameters):
    """ Creates a gcode text file
    Parameters
    ----------
    printpoints_dict: dict with compas_slicer.geometry.Printpoint instances.
        The keys of the dictionary are setup in the following way:
        {['layer_%d' % i] =
            {['path_%d' % i] =
                [printpoint_1 , ....., printpoint_n ]
            }
        }
    parameters : dict with gcode parameters.
        The defaults for those parameters are in the file compas_slicer.parameters.defaults_gcode.

    Returns
    ----------
    str, gcode text file
    """
    logger.info('Generating gcode')
    gcode = ''

    # get all the necessary parameters:
    z_hop = get_param(parameters, key='z_hop', defaults_type='gcode')
    extruder_temp = get_param(parameters, key='extruder_temperature', defaults_type='gcode')
    bed_temp = get_param(parameters, key='bed_temperature', defaults_type='gcode')
    # ....

    # Iterate through the printpoints_dict and add information to the gcode str
    for layer_key in printpoints_dict:
        for path_key in printpoints_dict[layer_key]:

            gcode += 'this is a command %.4f %.4f %.4f \n' % (z_hop, extruder_temp, bed_temp)  # just remove this line

            # .... gcode += 'command'
    # Thanks!

    return gcode



__all__ = ['gcode_default_param']


def gcode_default_param(key):
    """ Returns the default parameter with the specified key. """
    if key in default_parameters:
        return default_parameters[key]
    else:
        raise ValueError('The parameter with key : ' + str(key) +
                         ' does not exist in the defaults of gcode parameters. ')


default_parameters = \
    {
        # These parameters are not thought-through! Feel free to change them as you see fit
        'extruder_temperature': 200,
        'bed_temperature': 60,
        'z_hop': 0.5,
        # add more parameters as you see fit
    }
